/* Generated by Cython 0.29.21 */

#ifndef __PYX_HAVE__scipy___lib___test_deprecation_def
#define __PYX_HAVE__scipy___lib___test_deprecation_def

#include "Python.h"

#ifndef __PYX_HAVE_API__scipy___lib___test_deprecation_def

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C int __pyx_f_5scipy_4_lib_21_test_deprecation_def_foo_deprecated(void);
__PYX_EXTERN_C int __pyx_f_5scipy_4_lib_21_test_deprecation_def_foo(void);

#endif /* !__PYX_HAVE_API__scipy___lib___test_deprecation_def */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_test_deprecation_def(void);
#else
PyMODINIT_FUNC PyInit__test_deprecation_def(void);
#endif

#endif /* !__PYX_HAVE__scipy___lib___test_deprecation_def */
