/* Generated by Cython 0.29.21 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__scipy__special___ellip_harm_2
#define __PYX_HAVE_API__scipy__special___ellip_harm_2
/* Early includes */
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "sf_error.h"
#include "numpy/npy_math.h"
#include "lapack_defs.h"
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "_ellip_harm_2.pyx",
  "_ellip_harm.pxd",
  "__init__.pxd",
  "type.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);


/*--- Type declarations ---*/
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn;
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn;
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_in;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_in;
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn;

/* "cython_special.pxd":8
 *     double
 * 
 * cpdef number_t spherical_jn(long n, number_t z, bint derivative=*) nogil             # <<<<<<<<<<<<<<
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) nogil
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) nogil
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn {
  int __pyx_n;
  int derivative;
};

/* "cython_special.pxd":9
 * 
 * cpdef number_t spherical_jn(long n, number_t z, bint derivative=*) nogil
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) nogil             # <<<<<<<<<<<<<<
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) nogil
 * cpdef number_t spherical_kn(long n, number_t z, bint derivative=*) nogil
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn {
  int __pyx_n;
  int derivative;
};

/* "cython_special.pxd":10
 * cpdef number_t spherical_jn(long n, number_t z, bint derivative=*) nogil
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) nogil
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) nogil             # <<<<<<<<<<<<<<
 * cpdef number_t spherical_kn(long n, number_t z, bint derivative=*) nogil
 * 
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_in {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_in {
  int __pyx_n;
  int derivative;
};

/* "cython_special.pxd":11
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) nogil
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) nogil
 * cpdef number_t spherical_kn(long n, number_t z, bint derivative=*) nogil             # <<<<<<<<<<<<<<
 * 
 * ctypedef fused Dd_number_t:
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn {
  int __pyx_n;
  int derivative;
};

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t;
typedef struct __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t;

/* "scipy/special/_ellip_harm_2.pyx":12
 * from ._ellip_harm cimport ellip_harmonic, ellip_harm_eval, lame_coefficients
 * 
 * ctypedef struct _ellip_data_t:             # <<<<<<<<<<<<<<
 *     double *eval
 *     double h2, k2
 */
struct __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t {
  double *eval;
  double h2;
  double k2;
  int n;
  int p;
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_CBLAS_INT(CBLAS_INT value);

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE CBLAS_INT __Pyx_PyInt_As_CBLAS_INT(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cython' */

/* Module declarations from 'scipy.special.cython_special' */

/* Module declarations from 'scipy.special' */

/* Module declarations from 'scipy.special.sf_error' */

/* Module declarations from 'numpy.math' */

/* Module declarations from 'scipy.special._ellip_harm' */
static CYTHON_INLINE double *__pyx_f_5scipy_7special_11_ellip_harm_lame_coefficients(double, double, int, int, void **, double, double); /*proto*/
static CYTHON_INLINE double __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(double, double, int, int, double, double *, double, double); /*proto*/
static CYTHON_INLINE double __pyx_f_5scipy_7special_11_ellip_harm_ellip_harmonic(double, double, int, int, double, double, double); /*proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void); /*proto*/

/* Module declarations from 'scipy.special._ellip_harm_2' */
__PYX_EXTERN_C int wrap_PyUFunc_getfperr(void); /*proto*/
#define __Pyx_MODULE_NAME "scipy.special._ellip_harm_2"
extern int __pyx_module_is_main_scipy__special___ellip_harm_2;
int __pyx_module_is_main_scipy__special___ellip_harm_2 = 0;

/* Implementation of 'scipy.special._ellip_harm_2' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_[] = "*";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_h2[] = "h2";
static const char __pyx_k_k2[] = "k2";
static const char __pyx_k_alg[] = "alg";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_mod[] = "mod";
static const char __pyx_k_nan[] = "nan";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_eigv[] = "eigv";
static const char __pyx_k_err1[] = "err1";
static const char __pyx_k_err2[] = "err2";
static const char __pyx_k_err3[] = "err3";
static const char __pyx_k_eval[] = "eval";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_quad[] = "quad";
static const char __pyx_k_res1[] = "res1";
static const char __pyx_k_res2[] = "res2";
static const char __pyx_k_res3[] = "res3";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_wvar[] = "wvar";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ctypes[] = "ctypes";
static const char __pyx_k_epsabs[] = "epsabs";
static const char __pyx_k_epsrel[] = "epsrel";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_weight[] = "weight";
static const char __pyx_k_bufferp[] = "bufferp";
static const char __pyx_k_capsule[] = "capsule";
static const char __pyx_k_ellipsoid[] = "_ellipsoid";
static const char __pyx_k_F_integrand[] = "_F_integrand";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_from_cython[] = "from_cython";
static const char __pyx_k_F_integrand1[] = "_F_integrand1";
static const char __pyx_k_F_integrand2[] = "_F_integrand2";
static const char __pyx_k_F_integrand3[] = "_F_integrand3";
static const char __pyx_k_F_integrand4[] = "_F_integrand4";
static const char __pyx_k_ellipsoid_norm[] = "_ellipsoid_norm";
static const char __pyx_k_scipy_integrate[] = "scipy.integrate";
static const char __pyx_k_LowLevelCallable[] = "LowLevelCallable";
static const char __pyx_k_ellip_harm_2_pyx[] = "_ellip_harm_2.pyx";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_scipy__lib__ccallback[] = "scipy._lib._ccallback";
static const char __pyx_k_scipy_special__ellip_harm_2[] = "scipy.special._ellip_harm_2";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static PyObject *__pyx_n_s_;
static PyObject *__pyx_n_u_F_integrand;
static PyObject *__pyx_n_u_F_integrand1;
static PyObject *__pyx_n_u_F_integrand2;
static PyObject *__pyx_n_u_F_integrand3;
static PyObject *__pyx_n_u_F_integrand4;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_LowLevelCallable;
static PyObject *__pyx_n_u_alg;
static PyObject *__pyx_n_s_bufferp;
static PyObject *__pyx_n_s_capsule;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_ctypes;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_eigv;
static PyObject *__pyx_kp_s_ellip_harm_2_pyx;
static PyObject *__pyx_n_s_ellipsoid;
static PyObject *__pyx_n_s_ellipsoid_norm;
static PyObject *__pyx_n_s_epsabs;
static PyObject *__pyx_n_s_epsrel;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_err1;
static PyObject *__pyx_n_s_err2;
static PyObject *__pyx_n_s_err3;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_eval;
static PyObject *__pyx_n_s_from_cython;
static PyObject *__pyx_n_s_h;
static PyObject *__pyx_n_s_h2;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_k2;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mod;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_nan;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_quad;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_res1;
static PyObject *__pyx_n_s_res2;
static PyObject *__pyx_n_s_res3;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_scipy__lib__ccallback;
static PyObject *__pyx_n_s_scipy_integrate;
static PyObject *__pyx_n_s_scipy_special__ellip_harm_2;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_weight;
static PyObject *__pyx_n_s_wvar;
static PyObject *__pyx_pf_5scipy_7special_13_ellip_harm_2__ellipsoid(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_h2, double __pyx_v_k2, int __pyx_v_n, int __pyx_v_p, double __pyx_v_s); /* proto */
static PyObject *__pyx_pf_5scipy_7special_13_ellip_harm_2_2_ellipsoid_norm(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_h2, double __pyx_v_k2, int __pyx_v_n, int __pyx_v_p); /* proto */
static PyObject *__pyx_float_1eneg_15;
static PyObject *__pyx_float_1eneg_300;
static PyObject *__pyx_float_neg_0_5;
static PyObject *__pyx_int_0;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
/* Late includes */

/* "scipy/special/_ellip_harm_2.pyx":17
 *     int n, p
 * 
 * cdef double _F_integrand(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, result
 */

static double __pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand(double __pyx_v_t, void *__pyx_v_user_data) {
  __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *__pyx_v_data;
  double __pyx_v_h2;
  double __pyx_v_k2;
  double __pyx_v_t2;
  double __pyx_v_i;
  double __pyx_v_result;
  int __pyx_v_n;
  int __pyx_v_p;
  double *__pyx_v_eval;
  double __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  double *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "scipy/special/_ellip_harm_2.pyx":18
 * 
 * cdef double _F_integrand(double t, void *user_data) nogil:
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data             # <<<<<<<<<<<<<<
 *     cdef double h2, k2, t2, i, a, result
 *     cdef int n, p
 */
  __pyx_v_data = ((__pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *)__pyx_v_user_data);

  /* "scipy/special/_ellip_harm_2.pyx":22
 *     cdef int n, p
 *     cdef double * eval
 *     t2 = t*t             # <<<<<<<<<<<<<<
 *     h2 = data[0].h2
 *     k2 = data[0].k2
 */
  __pyx_v_t2 = (__pyx_v_t * __pyx_v_t);

  /* "scipy/special/_ellip_harm_2.pyx":23
 *     cdef double * eval
 *     t2 = t*t
 *     h2 = data[0].h2             # <<<<<<<<<<<<<<
 *     k2 = data[0].k2
 *     n = data[0].n
 */
  __pyx_t_1 = (__pyx_v_data[0]).h2;
  __pyx_v_h2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":24
 *     t2 = t*t
 *     h2 = data[0].h2
 *     k2 = data[0].k2             # <<<<<<<<<<<<<<
 *     n = data[0].n
 *     p = data[0].p
 */
  __pyx_t_1 = (__pyx_v_data[0]).k2;
  __pyx_v_k2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":25
 *     h2 = data[0].h2
 *     k2 = data[0].k2
 *     n = data[0].n             # <<<<<<<<<<<<<<
 *     p = data[0].p
 *     eval = data[0].eval
 */
  __pyx_t_2 = (__pyx_v_data[0]).n;
  __pyx_v_n = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":26
 *     k2 = data[0].k2
 *     n = data[0].n
 *     p = data[0].p             # <<<<<<<<<<<<<<
 *     eval = data[0].eval
 *     i = ellip_harm_eval( h2, k2, n, p, 1/t, eval, 1, 1)
 */
  __pyx_t_2 = (__pyx_v_data[0]).p;
  __pyx_v_p = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":27
 *     n = data[0].n
 *     p = data[0].p
 *     eval = data[0].eval             # <<<<<<<<<<<<<<
 *     i = ellip_harm_eval( h2, k2, n, p, 1/t, eval, 1, 1)
 *     result = 1/(i*i*sqrt(1 - t2*k2)*sqrt(1 - t2*h2))
 */
  __pyx_t_3 = (__pyx_v_data[0]).eval;
  __pyx_v_eval = __pyx_t_3;

  /* "scipy/special/_ellip_harm_2.pyx":28
 *     p = data[0].p
 *     eval = data[0].eval
 *     i = ellip_harm_eval( h2, k2, n, p, 1/t, eval, 1, 1)             # <<<<<<<<<<<<<<
 *     result = 1/(i*i*sqrt(1 - t2*k2)*sqrt(1 - t2*h2))
 *     return result
 */
  if (unlikely(__pyx_v_t == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 28, __pyx_L1_error)
  }
  __pyx_v_i = __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, (1.0 / __pyx_v_t), __pyx_v_eval, 1.0, 1.0);

  /* "scipy/special/_ellip_harm_2.pyx":29
 *     eval = data[0].eval
 *     i = ellip_harm_eval( h2, k2, n, p, 1/t, eval, 1, 1)
 *     result = 1/(i*i*sqrt(1 - t2*k2)*sqrt(1 - t2*h2))             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_1 = (((__pyx_v_i * __pyx_v_i) * sqrt((1.0 - (__pyx_v_t2 * __pyx_v_k2)))) * sqrt((1.0 - (__pyx_v_t2 * __pyx_v_h2))));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 29, __pyx_L1_error)
  }
  __pyx_v_result = (1.0 / __pyx_t_1);

  /* "scipy/special/_ellip_harm_2.pyx":30
 *     i = ellip_harm_eval( h2, k2, n, p, 1/t, eval, 1, 1)
 *     result = 1/(i*i*sqrt(1 - t2*k2)*sqrt(1 - t2*h2))
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef double _F_integrand1(double t, void *user_data) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":17
 *     int n, p
 * 
 * cdef double _F_integrand(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("scipy.special._ellip_harm_2._F_integrand", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "scipy/special/_ellip_harm_2.pyx":32
 *     return result
 * 
 * cdef double _F_integrand1(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

static double __pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand1(double __pyx_v_t, void *__pyx_v_user_data) {
  __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *__pyx_v_data;
  double __pyx_v_h2;
  double __pyx_v_k2;
  CYTHON_UNUSED double __pyx_v_t2;
  double __pyx_v_i;
  double __pyx_v_h;
  double __pyx_v_result;
  int __pyx_v_n;
  int __pyx_v_p;
  double *__pyx_v_eval;
  double __pyx_v_k;
  double __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  double *__pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "scipy/special/_ellip_harm_2.pyx":33
 * 
 * cdef double _F_integrand1(double t, void *user_data) nogil:
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data             # <<<<<<<<<<<<<<
 *     cdef double h2, k2, t2, i, a, h, result
 *     cdef int n, p
 */
  __pyx_v_data = ((__pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *)__pyx_v_user_data);

  /* "scipy/special/_ellip_harm_2.pyx":37
 *     cdef int n, p
 *     cdef double * eval
 *     t2 = t*t             # <<<<<<<<<<<<<<
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 */
  __pyx_v_t2 = (__pyx_v_t * __pyx_v_t);

  /* "scipy/special/_ellip_harm_2.pyx":38
 *     cdef double * eval
 *     t2 = t*t
 *     h2 = data[0].h2             # <<<<<<<<<<<<<<
 *     k2 =data[0].k2
 *     n = data[0].n
 */
  __pyx_t_1 = (__pyx_v_data[0]).h2;
  __pyx_v_h2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":39
 *     t2 = t*t
 *     h2 = data[0].h2
 *     k2 =data[0].k2             # <<<<<<<<<<<<<<
 *     n = data[0].n
 *     p = data[0].p
 */
  __pyx_t_1 = (__pyx_v_data[0]).k2;
  __pyx_v_k2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":40
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 *     n = data[0].n             # <<<<<<<<<<<<<<
 *     p = data[0].p
 *     eval = data[0].eval
 */
  __pyx_t_2 = (__pyx_v_data[0]).n;
  __pyx_v_n = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":41
 *     k2 =data[0].k2
 *     n = data[0].n
 *     p = data[0].p             # <<<<<<<<<<<<<<
 *     eval = data[0].eval
 * 
 */
  __pyx_t_2 = (__pyx_v_data[0]).p;
  __pyx_v_p = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":42
 *     n = data[0].n
 *     p = data[0].p
 *     eval = data[0].eval             # <<<<<<<<<<<<<<
 * 
 *     h = sqrt(h2)
 */
  __pyx_t_3 = (__pyx_v_data[0]).eval;
  __pyx_v_eval = __pyx_t_3;

  /* "scipy/special/_ellip_harm_2.pyx":44
 *     eval = data[0].eval
 * 
 *     h = sqrt(h2)             # <<<<<<<<<<<<<<
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 */
  __pyx_v_h = sqrt(__pyx_v_h2);

  /* "scipy/special/_ellip_harm_2.pyx":45
 * 
 *     h = sqrt(h2)
 *     k = sqrt(k2)             # <<<<<<<<<<<<<<
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i/sqrt((t + h)*(t + k))
 */
  __pyx_v_k = sqrt(__pyx_v_k2);

  /* "scipy/special/_ellip_harm_2.pyx":46
 *     h = sqrt(h2)
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)             # <<<<<<<<<<<<<<
 *     result = i*i/sqrt((t + h)*(t + k))
 *     return result
 */
  __pyx_v_i = __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, __pyx_v_t, __pyx_v_eval, 1.0, 1.0);

  /* "scipy/special/_ellip_harm_2.pyx":47
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i/sqrt((t + h)*(t + k))             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_1 = (__pyx_v_i * __pyx_v_i);
  __pyx_t_4 = sqrt(((__pyx_v_t + __pyx_v_h) * (__pyx_v_t + __pyx_v_k)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 47, __pyx_L1_error)
  }
  __pyx_v_result = (__pyx_t_1 / __pyx_t_4);

  /* "scipy/special/_ellip_harm_2.pyx":48
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i/sqrt((t + h)*(t + k))
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef double _F_integrand2(double t, void *user_data) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":32
 *     return result
 * 
 * cdef double _F_integrand1(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("scipy.special._ellip_harm_2._F_integrand1", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "scipy/special/_ellip_harm_2.pyx":50
 *     return result
 * 
 * cdef double _F_integrand2(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

static double __pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand2(double __pyx_v_t, void *__pyx_v_user_data) {
  __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *__pyx_v_data;
  double __pyx_v_h2;
  double __pyx_v_k2;
  double __pyx_v_t2;
  double __pyx_v_i;
  double __pyx_v_h;
  double __pyx_v_result;
  int __pyx_v_n;
  int __pyx_v_p;
  double *__pyx_v_eval;
  double __pyx_v_k;
  double __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  double *__pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "scipy/special/_ellip_harm_2.pyx":51
 * 
 * cdef double _F_integrand2(double t, void *user_data) nogil:
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data             # <<<<<<<<<<<<<<
 *     cdef double h2, k2, t2, i, a, h, result
 *     cdef int n, p
 */
  __pyx_v_data = ((__pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *)__pyx_v_user_data);

  /* "scipy/special/_ellip_harm_2.pyx":55
 *     cdef int n, p
 *     cdef double * eval
 *     t2 = t*t             # <<<<<<<<<<<<<<
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 */
  __pyx_v_t2 = (__pyx_v_t * __pyx_v_t);

  /* "scipy/special/_ellip_harm_2.pyx":56
 *     cdef double * eval
 *     t2 = t*t
 *     h2 = data[0].h2             # <<<<<<<<<<<<<<
 *     k2 =data[0].k2
 *     n = data[0].n
 */
  __pyx_t_1 = (__pyx_v_data[0]).h2;
  __pyx_v_h2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":57
 *     t2 = t*t
 *     h2 = data[0].h2
 *     k2 =data[0].k2             # <<<<<<<<<<<<<<
 *     n = data[0].n
 *     p = data[0].p
 */
  __pyx_t_1 = (__pyx_v_data[0]).k2;
  __pyx_v_k2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":58
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 *     n = data[0].n             # <<<<<<<<<<<<<<
 *     p = data[0].p
 *     eval = data[0].eval
 */
  __pyx_t_2 = (__pyx_v_data[0]).n;
  __pyx_v_n = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":59
 *     k2 =data[0].k2
 *     n = data[0].n
 *     p = data[0].p             # <<<<<<<<<<<<<<
 *     eval = data[0].eval
 * 
 */
  __pyx_t_2 = (__pyx_v_data[0]).p;
  __pyx_v_p = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":60
 *     n = data[0].n
 *     p = data[0].p
 *     eval = data[0].eval             # <<<<<<<<<<<<<<
 * 
 *     h = sqrt(h2)
 */
  __pyx_t_3 = (__pyx_v_data[0]).eval;
  __pyx_v_eval = __pyx_t_3;

  /* "scipy/special/_ellip_harm_2.pyx":62
 *     eval = data[0].eval
 * 
 *     h = sqrt(h2)             # <<<<<<<<<<<<<<
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 */
  __pyx_v_h = sqrt(__pyx_v_h2);

  /* "scipy/special/_ellip_harm_2.pyx":63
 * 
 *     h = sqrt(h2)
 *     k = sqrt(k2)             # <<<<<<<<<<<<<<
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = t2*i*i/sqrt((t + h)*(t + k))
 */
  __pyx_v_k = sqrt(__pyx_v_k2);

  /* "scipy/special/_ellip_harm_2.pyx":64
 *     h = sqrt(h2)
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)             # <<<<<<<<<<<<<<
 *     result = t2*i*i/sqrt((t + h)*(t + k))
 *     return result
 */
  __pyx_v_i = __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, __pyx_v_t, __pyx_v_eval, 1.0, 1.0);

  /* "scipy/special/_ellip_harm_2.pyx":65
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = t2*i*i/sqrt((t + h)*(t + k))             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_1 = ((__pyx_v_t2 * __pyx_v_i) * __pyx_v_i);
  __pyx_t_4 = sqrt(((__pyx_v_t + __pyx_v_h) * (__pyx_v_t + __pyx_v_k)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 65, __pyx_L1_error)
  }
  __pyx_v_result = (__pyx_t_1 / __pyx_t_4);

  /* "scipy/special/_ellip_harm_2.pyx":66
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = t2*i*i/sqrt((t + h)*(t + k))
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef double _F_integrand3(double t, void *user_data) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":50
 *     return result
 * 
 * cdef double _F_integrand2(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("scipy.special._ellip_harm_2._F_integrand2", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "scipy/special/_ellip_harm_2.pyx":68
 *     return result
 * 
 * cdef double _F_integrand3(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

static double __pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand3(double __pyx_v_t, void *__pyx_v_user_data) {
  __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *__pyx_v_data;
  double __pyx_v_h2;
  double __pyx_v_k2;
  double __pyx_v_t2;
  double __pyx_v_i;
  double __pyx_v_h;
  double __pyx_v_result;
  int __pyx_v_n;
  int __pyx_v_p;
  double *__pyx_v_eval;
  CYTHON_UNUSED double __pyx_v_k;
  double __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  double *__pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "scipy/special/_ellip_harm_2.pyx":69
 * 
 * cdef double _F_integrand3(double t, void *user_data) nogil:
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data             # <<<<<<<<<<<<<<
 *     cdef double h2, k2, t2, i, a, h, result
 *     cdef int n, p
 */
  __pyx_v_data = ((__pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *)__pyx_v_user_data);

  /* "scipy/special/_ellip_harm_2.pyx":73
 *     cdef int n, p
 *     cdef double * eval
 *     t2 = t*t             # <<<<<<<<<<<<<<
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 */
  __pyx_v_t2 = (__pyx_v_t * __pyx_v_t);

  /* "scipy/special/_ellip_harm_2.pyx":74
 *     cdef double * eval
 *     t2 = t*t
 *     h2 = data[0].h2             # <<<<<<<<<<<<<<
 *     k2 =data[0].k2
 *     n = data[0].n
 */
  __pyx_t_1 = (__pyx_v_data[0]).h2;
  __pyx_v_h2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":75
 *     t2 = t*t
 *     h2 = data[0].h2
 *     k2 =data[0].k2             # <<<<<<<<<<<<<<
 *     n = data[0].n
 *     p = data[0].p
 */
  __pyx_t_1 = (__pyx_v_data[0]).k2;
  __pyx_v_k2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":76
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 *     n = data[0].n             # <<<<<<<<<<<<<<
 *     p = data[0].p
 *     eval = data[0].eval
 */
  __pyx_t_2 = (__pyx_v_data[0]).n;
  __pyx_v_n = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":77
 *     k2 =data[0].k2
 *     n = data[0].n
 *     p = data[0].p             # <<<<<<<<<<<<<<
 *     eval = data[0].eval
 * 
 */
  __pyx_t_2 = (__pyx_v_data[0]).p;
  __pyx_v_p = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":78
 *     n = data[0].n
 *     p = data[0].p
 *     eval = data[0].eval             # <<<<<<<<<<<<<<
 * 
 *     h = sqrt(h2)
 */
  __pyx_t_3 = (__pyx_v_data[0]).eval;
  __pyx_v_eval = __pyx_t_3;

  /* "scipy/special/_ellip_harm_2.pyx":80
 *     eval = data[0].eval
 * 
 *     h = sqrt(h2)             # <<<<<<<<<<<<<<
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 */
  __pyx_v_h = sqrt(__pyx_v_h2);

  /* "scipy/special/_ellip_harm_2.pyx":81
 * 
 *     h = sqrt(h2)
 *     k = sqrt(k2)             # <<<<<<<<<<<<<<
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i/sqrt((t + h)*(k2 - t2))
 */
  __pyx_v_k = sqrt(__pyx_v_k2);

  /* "scipy/special/_ellip_harm_2.pyx":82
 *     h = sqrt(h2)
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)             # <<<<<<<<<<<<<<
 *     result = i*i/sqrt((t + h)*(k2 - t2))
 *     return result
 */
  __pyx_v_i = __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, __pyx_v_t, __pyx_v_eval, 1.0, 1.0);

  /* "scipy/special/_ellip_harm_2.pyx":83
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i/sqrt((t + h)*(k2 - t2))             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_1 = (__pyx_v_i * __pyx_v_i);
  __pyx_t_4 = sqrt(((__pyx_v_t + __pyx_v_h) * (__pyx_v_k2 - __pyx_v_t2)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 83, __pyx_L1_error)
  }
  __pyx_v_result = (__pyx_t_1 / __pyx_t_4);

  /* "scipy/special/_ellip_harm_2.pyx":84
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i/sqrt((t + h)*(k2 - t2))
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef double _F_integrand4(double t, void *user_data) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":68
 *     return result
 * 
 * cdef double _F_integrand3(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("scipy.special._ellip_harm_2._F_integrand3", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "scipy/special/_ellip_harm_2.pyx":86
 *     return result
 * 
 * cdef double _F_integrand4(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

static double __pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand4(double __pyx_v_t, void *__pyx_v_user_data) {
  __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *__pyx_v_data;
  double __pyx_v_h2;
  double __pyx_v_k2;
  double __pyx_v_t2;
  double __pyx_v_i;
  double __pyx_v_h;
  double __pyx_v_result;
  int __pyx_v_n;
  int __pyx_v_p;
  double *__pyx_v_eval;
  CYTHON_UNUSED double __pyx_v_k;
  double __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  double *__pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "scipy/special/_ellip_harm_2.pyx":87
 * 
 * cdef double _F_integrand4(double t, void *user_data) nogil:
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data             # <<<<<<<<<<<<<<
 *     cdef double h2, k2, t2, i, a, h, result
 *     cdef int n, p
 */
  __pyx_v_data = ((__pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t *)__pyx_v_user_data);

  /* "scipy/special/_ellip_harm_2.pyx":91
 *     cdef int n, p
 *     cdef double *eval
 *     t2 = t*t             # <<<<<<<<<<<<<<
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 */
  __pyx_v_t2 = (__pyx_v_t * __pyx_v_t);

  /* "scipy/special/_ellip_harm_2.pyx":92
 *     cdef double *eval
 *     t2 = t*t
 *     h2 = data[0].h2             # <<<<<<<<<<<<<<
 *     k2 =data[0].k2
 *     n = data[0].n
 */
  __pyx_t_1 = (__pyx_v_data[0]).h2;
  __pyx_v_h2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":93
 *     t2 = t*t
 *     h2 = data[0].h2
 *     k2 =data[0].k2             # <<<<<<<<<<<<<<
 *     n = data[0].n
 *     p = data[0].p
 */
  __pyx_t_1 = (__pyx_v_data[0]).k2;
  __pyx_v_k2 = __pyx_t_1;

  /* "scipy/special/_ellip_harm_2.pyx":94
 *     h2 = data[0].h2
 *     k2 =data[0].k2
 *     n = data[0].n             # <<<<<<<<<<<<<<
 *     p = data[0].p
 *     eval = data[0].eval
 */
  __pyx_t_2 = (__pyx_v_data[0]).n;
  __pyx_v_n = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":95
 *     k2 =data[0].k2
 *     n = data[0].n
 *     p = data[0].p             # <<<<<<<<<<<<<<
 *     eval = data[0].eval
 * 
 */
  __pyx_t_2 = (__pyx_v_data[0]).p;
  __pyx_v_p = __pyx_t_2;

  /* "scipy/special/_ellip_harm_2.pyx":96
 *     n = data[0].n
 *     p = data[0].p
 *     eval = data[0].eval             # <<<<<<<<<<<<<<
 * 
 *     h = sqrt(h2)
 */
  __pyx_t_3 = (__pyx_v_data[0]).eval;
  __pyx_v_eval = __pyx_t_3;

  /* "scipy/special/_ellip_harm_2.pyx":98
 *     eval = data[0].eval
 * 
 *     h = sqrt(h2)             # <<<<<<<<<<<<<<
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 */
  __pyx_v_h = sqrt(__pyx_v_h2);

  /* "scipy/special/_ellip_harm_2.pyx":99
 * 
 *     h = sqrt(h2)
 *     k = sqrt(k2)             # <<<<<<<<<<<<<<
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i*t2/sqrt((t + h)*(k2 - t2))
 */
  __pyx_v_k = sqrt(__pyx_v_k2);

  /* "scipy/special/_ellip_harm_2.pyx":100
 *     h = sqrt(h2)
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)             # <<<<<<<<<<<<<<
 *     result = i*i*t2/sqrt((t + h)*(k2 - t2))
 *     return result
 */
  __pyx_v_i = __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, __pyx_v_t, __pyx_v_eval, 1.0, 1.0);

  /* "scipy/special/_ellip_harm_2.pyx":101
 *     k = sqrt(k2)
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i*t2/sqrt((t + h)*(k2 - t2))             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_1 = ((__pyx_v_i * __pyx_v_i) * __pyx_v_t2);
  __pyx_t_4 = sqrt(((__pyx_v_t + __pyx_v_h) * (__pyx_v_k2 - __pyx_v_t2)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 101, __pyx_L1_error)
  }
  __pyx_v_result = (__pyx_t_1 / __pyx_t_4);

  /* "scipy/special/_ellip_harm_2.pyx":102
 *     i = ellip_harm_eval( h2, k2, n, p, t, eval, 1, 1)
 *     result = i*i*t2/sqrt((t + h)*(k2 - t2))
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":86
 *     return result
 * 
 * cdef double _F_integrand4(double t, void *user_data) nogil:             # <<<<<<<<<<<<<<
 *     cdef _ellip_data_t *data = <_ellip_data_t *>user_data
 *     cdef double h2, k2, t2, i, a, h, result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("scipy.special._ellip_harm_2._F_integrand4", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "scipy/special/_ellip_harm_2.pyx":105
 * 
 * 
 * def _ellipsoid(double h2, double k2, int n, int p, double s):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */

/* Python wrapper */
static PyObject *__pyx_pw_5scipy_7special_13_ellip_harm_2_1_ellipsoid(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5scipy_7special_13_ellip_harm_2_1_ellipsoid = {"_ellipsoid", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5scipy_7special_13_ellip_harm_2_1_ellipsoid, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5scipy_7special_13_ellip_harm_2_1_ellipsoid(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_h2;
  double __pyx_v_k2;
  int __pyx_v_n;
  int __pyx_v_p;
  double __pyx_v_s;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_ellipsoid (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_h2,&__pyx_n_s_k2,&__pyx_n_s_n,&__pyx_n_s_p,&__pyx_n_s_s,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_ellipsoid", 1, 5, 5, 1); __PYX_ERR(0, 105, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_ellipsoid", 1, 5, 5, 2); __PYX_ERR(0, 105, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_ellipsoid", 1, 5, 5, 3); __PYX_ERR(0, 105, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_ellipsoid", 1, 5, 5, 4); __PYX_ERR(0, 105, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_ellipsoid") < 0)) __PYX_ERR(0, 105, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_h2 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_h2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
    __pyx_v_k2 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_k2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_p == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
    __pyx_v_s = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_s == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_ellipsoid", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 105, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("scipy.special._ellip_harm_2._ellipsoid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5scipy_7special_13_ellip_harm_2__ellipsoid(__pyx_self, __pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, __pyx_v_s);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5scipy_7special_13_ellip_harm_2__ellipsoid(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_h2, double __pyx_v_k2, int __pyx_v_n, int __pyx_v_p, double __pyx_v_s) {
  PyObject *__pyx_v_mod = NULL;
  PyObject *__pyx_v_quad = NULL;
  __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t __pyx_v_data;
  double *__pyx_v_eval;
  void *__pyx_v_bufferp;
  double __pyx_v_res;
  double __pyx_v_err;
  PyObject *__pyx_v_capsule = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  double __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ellipsoid", 0);

  /* "scipy/special/_ellip_harm_2.pyx":106
 * 
 * def _ellipsoid(double h2, double k2, int n, int p, double s):
 *     import scipy.special._ellip_harm_2 as mod             # <<<<<<<<<<<<<<
 *     from scipy.integrate import quad
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_);
  __Pyx_GIVEREF(__pyx_n_s_);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scipy_special__ellip_harm_2, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mod = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":107
 * def _ellipsoid(double h2, double k2, int n, int p, double s):
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad             # <<<<<<<<<<<<<<
 * 
 *     cdef _ellip_data_t data
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_quad);
  __Pyx_GIVEREF(__pyx_n_s_quad);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_quad);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_scipy_integrate, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_quad); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_quad = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":113
 *     cdef double * eval
 *     cdef void *bufferp
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)             # <<<<<<<<<<<<<<
 *     if not eval:
 *         return nan
 */
  __pyx_v_eval = __pyx_f_5scipy_7special_11_ellip_harm_lame_coefficients(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, (&__pyx_v_bufferp), 1.0, 1.0);

  /* "scipy/special/_ellip_harm_2.pyx":114
 *     cdef void *bufferp
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)
 *     if not eval:             # <<<<<<<<<<<<<<
 *         return nan
 * 
 */
  __pyx_t_3 = ((!(__pyx_v_eval != 0)) != 0);
  if (__pyx_t_3) {

    /* "scipy/special/_ellip_harm_2.pyx":115
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)
 *     if not eval:
 *         return nan             # <<<<<<<<<<<<<<
 * 
 *     data.h2 = h2
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_nan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "scipy/special/_ellip_harm_2.pyx":114
 *     cdef void *bufferp
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)
 *     if not eval:             # <<<<<<<<<<<<<<
 *         return nan
 * 
 */
  }

  /* "scipy/special/_ellip_harm_2.pyx":117
 *         return nan
 * 
 *     data.h2 = h2             # <<<<<<<<<<<<<<
 *     data.k2 = k2
 *     data.n = n
 */
  __pyx_v_data.h2 = __pyx_v_h2;

  /* "scipy/special/_ellip_harm_2.pyx":118
 * 
 *     data.h2 = h2
 *     data.k2 = k2             # <<<<<<<<<<<<<<
 *     data.n = n
 *     data.p = p
 */
  __pyx_v_data.k2 = __pyx_v_k2;

  /* "scipy/special/_ellip_harm_2.pyx":119
 *     data.h2 = h2
 *     data.k2 = k2
 *     data.n = n             # <<<<<<<<<<<<<<
 *     data.p = p
 *     data.eval = eval
 */
  __pyx_v_data.n = __pyx_v_n;

  /* "scipy/special/_ellip_harm_2.pyx":120
 *     data.k2 = k2
 *     data.n = n
 *     data.p = p             # <<<<<<<<<<<<<<
 *     data.eval = eval
 * 
 */
  __pyx_v_data.p = __pyx_v_p;

  /* "scipy/special/_ellip_harm_2.pyx":121
 *     data.n = n
 *     data.p = p
 *     data.eval = eval             # <<<<<<<<<<<<<<
 * 
 *     cdef double res, err
 */
  __pyx_v_data.eval = __pyx_v_eval;

  /* "scipy/special/_ellip_harm_2.pyx":125
 *     cdef double res, err
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand", capsule), 0, 1/s,
 */
  /*try:*/ {

    /* "scipy/special/_ellip_harm_2.pyx":126
 * 
 *     try:
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)             # <<<<<<<<<<<<<<
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand", capsule), 0, 1/s,
 *                                                      epsabs=1e-300, epsrel=1e-15)
 */
    __pyx_t_1 = PyCapsule_New(((void *)(&__pyx_v_data)), NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_capsule = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "scipy/special/_ellip_harm_2.pyx":127
 *     try:
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand", capsule), 0, 1/s,             # <<<<<<<<<<<<<<
 *                                                      epsabs=1e-300, epsrel=1e-15)
 *     finally:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LowLevelCallable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_from_cython); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand, __pyx_v_capsule};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand, __pyx_v_capsule};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_mod);
      __Pyx_GIVEREF(__pyx_v_mod);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_mod);
      __Pyx_INCREF(__pyx_n_u_F_integrand);
      __Pyx_GIVEREF(__pyx_n_u_F_integrand);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_n_u_F_integrand);
      __Pyx_INCREF(__pyx_v_capsule);
      __Pyx_GIVEREF(__pyx_v_capsule);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_capsule);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__pyx_v_s == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 127, __pyx_L5_error)
    }
    __pyx_t_4 = PyFloat_FromDouble((1.0 / __pyx_v_s)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;

    /* "scipy/special/_ellip_harm_2.pyx":128
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand", capsule), 0, 1/s,
 *                                                      epsabs=1e-300, epsrel=1e-15)             # <<<<<<<<<<<<<<
 *     finally:
 *         free(bufferp)
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 128, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_epsabs, __pyx_float_1eneg_300) < 0) __PYX_ERR(0, 128, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_epsrel, __pyx_float_1eneg_15) < 0) __PYX_ERR(0, 128, __pyx_L5_error)

    /* "scipy/special/_ellip_harm_2.pyx":127
 *     try:
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand", capsule), 0, 1/s,             # <<<<<<<<<<<<<<
 *                                                      epsabs=1e-300, epsrel=1e-15)
 *     finally:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_quad, __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 127, __pyx_L5_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_4)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_6 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_2), 2) < 0) __PYX_ERR(0, 127, __pyx_L5_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 127, __pyx_L5_error)
      __pyx_L8_unpacking_done:;
    }
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_res = __pyx_t_8;
    __pyx_v_err = __pyx_t_9;
  }

  /* "scipy/special/_ellip_harm_2.pyx":130
 *                                                      epsabs=1e-300, epsrel=1e-15)
 *     finally:
 *         free(bufferp)             # <<<<<<<<<<<<<<
 *     if err > 1e-10*fabs(res) + 1e-290:
 *         return nan
 */
  /*finally:*/ {
    /*normal exit:*/{
      free(__pyx_v_bufferp);
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_5 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
      {
        free(__pyx_v_bufferp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "scipy/special/_ellip_harm_2.pyx":131
 *     finally:
 *         free(bufferp)
 *     if err > 1e-10*fabs(res) + 1e-290:             # <<<<<<<<<<<<<<
 *         return nan
 *     res = res*(2*n + 1)*ellip_harmonic( h2, k2, n, p, s, 1, 1)
 */
  __pyx_t_3 = ((__pyx_v_err > ((1e-10 * fabs(__pyx_v_res)) + 1e-290)) != 0);
  if (__pyx_t_3) {

    /* "scipy/special/_ellip_harm_2.pyx":132
 *         free(bufferp)
 *     if err > 1e-10*fabs(res) + 1e-290:
 *         return nan             # <<<<<<<<<<<<<<
 *     res = res*(2*n + 1)*ellip_harmonic( h2, k2, n, p, s, 1, 1)
 *     return res
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_nan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "scipy/special/_ellip_harm_2.pyx":131
 *     finally:
 *         free(bufferp)
 *     if err > 1e-10*fabs(res) + 1e-290:             # <<<<<<<<<<<<<<
 *         return nan
 *     res = res*(2*n + 1)*ellip_harmonic( h2, k2, n, p, s, 1, 1)
 */
  }

  /* "scipy/special/_ellip_harm_2.pyx":133
 *     if err > 1e-10*fabs(res) + 1e-290:
 *         return nan
 *     res = res*(2*n + 1)*ellip_harmonic( h2, k2, n, p, s, 1, 1)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  __pyx_v_res = ((__pyx_v_res * ((2 * __pyx_v_n) + 1)) * __pyx_f_5scipy_7special_11_ellip_harm_ellip_harmonic(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, __pyx_v_s, 1.0, 1.0));

  /* "scipy/special/_ellip_harm_2.pyx":134
 *         return nan
 *     res = res*(2*n + 1)*ellip_harmonic( h2, k2, n, p, s, 1, 1)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_res); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":105
 * 
 * 
 * def _ellipsoid(double h2, double k2, int n, int p, double s):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("scipy.special._ellip_harm_2._ellipsoid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mod);
  __Pyx_XDECREF(__pyx_v_quad);
  __Pyx_XDECREF(__pyx_v_capsule);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "scipy/special/_ellip_harm_2.pyx":137
 * 
 * 
 * def _ellipsoid_norm(double h2, double k2, int n, int p):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */

/* Python wrapper */
static PyObject *__pyx_pw_5scipy_7special_13_ellip_harm_2_3_ellipsoid_norm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5scipy_7special_13_ellip_harm_2_3_ellipsoid_norm = {"_ellipsoid_norm", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5scipy_7special_13_ellip_harm_2_3_ellipsoid_norm, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5scipy_7special_13_ellip_harm_2_3_ellipsoid_norm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_h2;
  double __pyx_v_k2;
  int __pyx_v_n;
  int __pyx_v_p;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_ellipsoid_norm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_h2,&__pyx_n_s_k2,&__pyx_n_s_n,&__pyx_n_s_p,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_ellipsoid_norm", 1, 4, 4, 1); __PYX_ERR(0, 137, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_ellipsoid_norm", 1, 4, 4, 2); __PYX_ERR(0, 137, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_ellipsoid_norm", 1, 4, 4, 3); __PYX_ERR(0, 137, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_ellipsoid_norm") < 0)) __PYX_ERR(0, 137, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_h2 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_h2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 137, __pyx_L3_error)
    __pyx_v_k2 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_k2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 137, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 137, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_p == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 137, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_ellipsoid_norm", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 137, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("scipy.special._ellip_harm_2._ellipsoid_norm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5scipy_7special_13_ellip_harm_2_2_ellipsoid_norm(__pyx_self, __pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5scipy_7special_13_ellip_harm_2_2_ellipsoid_norm(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_h2, double __pyx_v_k2, int __pyx_v_n, int __pyx_v_p) {
  PyObject *__pyx_v_mod = NULL;
  PyObject *__pyx_v_quad = NULL;
  __pyx_t_5scipy_7special_13_ellip_harm_2__ellip_data_t __pyx_v_data;
  void *__pyx_v_bufferp;
  double *__pyx_v_eval;
  double __pyx_v_res;
  double __pyx_v_res1;
  double __pyx_v_res2;
  double __pyx_v_res3;
  double __pyx_v_err;
  double __pyx_v_err1;
  double __pyx_v_err2;
  double __pyx_v_err3;
  double __pyx_v_h;
  double __pyx_v_k;
  PyObject *__pyx_v_capsule = NULL;
  PyObject *__pyx_v_wvar = NULL;
  double __pyx_v_error;
  double __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  double __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ellipsoid_norm", 0);

  /* "scipy/special/_ellip_harm_2.pyx":138
 * 
 * def _ellipsoid_norm(double h2, double k2, int n, int p):
 *     import scipy.special._ellip_harm_2 as mod             # <<<<<<<<<<<<<<
 *     from scipy.integrate import quad
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_);
  __Pyx_GIVEREF(__pyx_n_s_);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scipy_special__ellip_harm_2, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mod = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":139
 * def _ellipsoid_norm(double h2, double k2, int n, int p):
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad             # <<<<<<<<<<<<<<
 * 
 *     cdef _ellip_data_t data
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_quad);
  __Pyx_GIVEREF(__pyx_n_s_quad);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_quad);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_scipy_integrate, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_quad); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_quad = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":145
 *     cdef double *eigv
 *     cdef void *bufferp
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)             # <<<<<<<<<<<<<<
 *     if not eval:
 *         return nan
 */
  __pyx_v_eval = __pyx_f_5scipy_7special_11_ellip_harm_lame_coefficients(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, (&__pyx_v_bufferp), 1.0, 1.0);

  /* "scipy/special/_ellip_harm_2.pyx":146
 *     cdef void *bufferp
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)
 *     if not eval:             # <<<<<<<<<<<<<<
 *         return nan
 * 
 */
  __pyx_t_3 = ((!(__pyx_v_eval != 0)) != 0);
  if (__pyx_t_3) {

    /* "scipy/special/_ellip_harm_2.pyx":147
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)
 *     if not eval:
 *         return nan             # <<<<<<<<<<<<<<
 * 
 *     data.h2 = h2
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_nan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "scipy/special/_ellip_harm_2.pyx":146
 *     cdef void *bufferp
 *     eval = lame_coefficients(h2, k2, n, p, &bufferp, 1, 1)
 *     if not eval:             # <<<<<<<<<<<<<<
 *         return nan
 * 
 */
  }

  /* "scipy/special/_ellip_harm_2.pyx":149
 *         return nan
 * 
 *     data.h2 = h2             # <<<<<<<<<<<<<<
 *     data.k2 = k2
 *     data.n = n
 */
  __pyx_v_data.h2 = __pyx_v_h2;

  /* "scipy/special/_ellip_harm_2.pyx":150
 * 
 *     data.h2 = h2
 *     data.k2 = k2             # <<<<<<<<<<<<<<
 *     data.n = n
 *     data.p = p
 */
  __pyx_v_data.k2 = __pyx_v_k2;

  /* "scipy/special/_ellip_harm_2.pyx":151
 *     data.h2 = h2
 *     data.k2 = k2
 *     data.n = n             # <<<<<<<<<<<<<<
 *     data.p = p
 *     data.eval = eval
 */
  __pyx_v_data.n = __pyx_v_n;

  /* "scipy/special/_ellip_harm_2.pyx":152
 *     data.k2 = k2
 *     data.n = n
 *     data.p = p             # <<<<<<<<<<<<<<
 *     data.eval = eval
 * 
 */
  __pyx_v_data.p = __pyx_v_p;

  /* "scipy/special/_ellip_harm_2.pyx":153
 *     data.n = n
 *     data.p = p
 *     data.eval = eval             # <<<<<<<<<<<<<<
 * 
 *     cdef double res, res1, res2, res3, err, err1, err2, err3
 */
  __pyx_v_data.eval = __pyx_v_eval;

  /* "scipy/special/_ellip_harm_2.pyx":157
 *     cdef double res, res1, res2, res3, err, err1, err2, err3
 * 
 *     h = sqrt(h2)             # <<<<<<<<<<<<<<
 *     k = sqrt(k2)
 *     try:
 */
  __pyx_v_h = sqrt(__pyx_v_h2);

  /* "scipy/special/_ellip_harm_2.pyx":158
 * 
 *     h = sqrt(h2)
 *     k = sqrt(k2)             # <<<<<<<<<<<<<<
 *     try:
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 */
  __pyx_v_k = sqrt(__pyx_v_k2);

  /* "scipy/special/_ellip_harm_2.pyx":159
 *     h = sqrt(h2)
 *     k = sqrt(k2)
 *     try:             # <<<<<<<<<<<<<<
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 * 
 */
  /*try:*/ {

    /* "scipy/special/_ellip_harm_2.pyx":160
 *     k = sqrt(k2)
 *     try:
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)             # <<<<<<<<<<<<<<
 * 
 *         wvar = (-0.5, -0.5)
 */
    __pyx_t_1 = PyCapsule_New(((void *)(&__pyx_v_data)), NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_capsule = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "scipy/special/_ellip_harm_2.pyx":162
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 * 
 *         wvar = (-0.5, -0.5)             # <<<<<<<<<<<<<<
 * 
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand1", capsule), h, k,
 */
    __Pyx_INCREF(__pyx_tuple__2);
    __pyx_v_wvar = __pyx_tuple__2;

    /* "scipy/special/_ellip_harm_2.pyx":164
 *         wvar = (-0.5, -0.5)
 * 
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand1", capsule), h, k,             # <<<<<<<<<<<<<<
 *                         epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LowLevelCallable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_from_cython); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand1, __pyx_v_capsule};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand1, __pyx_v_capsule};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 164, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_mod);
      __Pyx_GIVEREF(__pyx_v_mod);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_mod);
      __Pyx_INCREF(__pyx_n_u_F_integrand1);
      __Pyx_GIVEREF(__pyx_n_u_F_integrand1);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_n_u_F_integrand1);
      __Pyx_INCREF(__pyx_v_capsule);
      __Pyx_GIVEREF(__pyx_v_capsule);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_capsule);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyFloat_FromDouble(__pyx_v_h); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyFloat_FromDouble(__pyx_v_k); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_6);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;

    /* "scipy/special/_ellip_harm_2.pyx":165
 * 
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand1", capsule), h, k,
 *                         epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)             # <<<<<<<<<<<<<<
 * 
 *         res1, err1 = quad(LowLevelCallable.from_cython(mod, "_F_integrand2", capsule), h, k,
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 165, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_epsabs, __pyx_float_1eneg_300) < 0) __PYX_ERR(0, 165, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_epsrel, __pyx_float_1eneg_15) < 0) __PYX_ERR(0, 165, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_weight, __pyx_n_u_alg) < 0) __PYX_ERR(0, 165, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_wvar, __pyx_v_wvar) < 0) __PYX_ERR(0, 165, __pyx_L5_error)

    /* "scipy/special/_ellip_harm_2.pyx":164
 *         wvar = (-0.5, -0.5)
 * 
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand1", capsule), h, k,             # <<<<<<<<<<<<<<
 *                         epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_quad, __pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 164, __pyx_L5_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 164, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_7(__pyx_t_1); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_2 = __pyx_t_7(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_1), 2) < 0) __PYX_ERR(0, 164, __pyx_L5_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 164, __pyx_L5_error)
      __pyx_L8_unpacking_done:;
    }
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 164, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_res = __pyx_t_8;
    __pyx_v_err = __pyx_t_9;

    /* "scipy/special/_ellip_harm_2.pyx":167
 *                         epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 *         res1, err1 = quad(LowLevelCallable.from_cython(mod, "_F_integrand2", capsule), h, k,             # <<<<<<<<<<<<<<
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LowLevelCallable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_from_cython); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand2, __pyx_v_capsule};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand2, __pyx_v_capsule};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_1 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_mod);
      __Pyx_GIVEREF(__pyx_v_mod);
      PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_mod);
      __Pyx_INCREF(__pyx_n_u_F_integrand2);
      __Pyx_GIVEREF(__pyx_n_u_F_integrand2);
      PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_n_u_F_integrand2);
      __Pyx_INCREF(__pyx_v_capsule);
      __Pyx_GIVEREF(__pyx_v_capsule);
      PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_5, __pyx_v_capsule);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyFloat_FromDouble(__pyx_v_h); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;

    /* "scipy/special/_ellip_harm_2.pyx":168
 * 
 *         res1, err1 = quad(LowLevelCallable.from_cython(mod, "_F_integrand2", capsule), h, k,
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)             # <<<<<<<<<<<<<<
 * 
 *         wvar = (0, -0.5)
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_epsabs, __pyx_float_1eneg_300) < 0) __PYX_ERR(0, 168, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_epsrel, __pyx_float_1eneg_15) < 0) __PYX_ERR(0, 168, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_weight, __pyx_n_u_alg) < 0) __PYX_ERR(0, 168, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_wvar, __pyx_v_wvar) < 0) __PYX_ERR(0, 168, __pyx_L5_error)

    /* "scipy/special/_ellip_harm_2.pyx":167
 *                         epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 *         res1, err1 = quad(LowLevelCallable.from_cython(mod, "_F_integrand2", capsule), h, k,             # <<<<<<<<<<<<<<
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_v_quad, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 167, __pyx_L5_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_2 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 2) < 0) __PYX_ERR(0, 167, __pyx_L5_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 167, __pyx_L5_error)
      __pyx_L10_unpacking_done:;
    }
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 167, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_res1 = __pyx_t_9;
    __pyx_v_err1 = __pyx_t_8;

    /* "scipy/special/_ellip_harm_2.pyx":170
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 *         wvar = (0, -0.5)             # <<<<<<<<<<<<<<
 * 
 *         res2, err2 = quad(LowLevelCallable.from_cython(mod, "_F_integrand3", capsule), 0, h,
 */
    __Pyx_INCREF(__pyx_tuple__3);
    __Pyx_DECREF_SET(__pyx_v_wvar, __pyx_tuple__3);

    /* "scipy/special/_ellip_harm_2.pyx":172
 *         wvar = (0, -0.5)
 * 
 *         res2, err2 = quad(LowLevelCallable.from_cython(mod, "_F_integrand3", capsule), 0, h,             # <<<<<<<<<<<<<<
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LowLevelCallable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_from_cython); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand3, __pyx_v_capsule};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mod, __pyx_n_u_F_integrand3, __pyx_v_capsule};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_mod);
      __Pyx_GIVEREF(__pyx_v_mod);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_mod);
      __Pyx_INCREF(__pyx_n_u_F_integrand3);
      __Pyx_GIVEREF(__pyx_n_u_F_integrand3);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_n_u_F_integrand3);
      __Pyx_INCREF(__pyx_v_capsule);
      __Pyx_GIVEREF(__pyx_v_capsule);
      PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_v_capsule);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;

    /* "scipy/special/_ellip_harm_2.pyx":173
 * 
 *         res2, err2 = quad(LowLevelCallable.from_cython(mod, "_F_integrand3", capsule), 0, h,
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)             # <<<<<<<<<<<<<<
 * 
 *         res3, err3 = quad(LowLevelCallable.from_cython(mod, "_F_integrand4", capsule), 0, h,
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_epsabs, __pyx_float_1eneg_300) < 0) __PYX_ERR(0, 173, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_epsrel, __pyx_float_1eneg_15) < 0) __PYX_ERR(0, 173, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_weight, __pyx_n_u_alg) < 0) __PYX_ERR(0, 173, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_wvar, __pyx_v_wvar) < 0) __PYX_ERR(0, 173, __pyx_L5_error)

    /* "scipy/special/_ellip_harm_2.pyx":172
 *         wvar = (0, -0.5)
 * 
 *         res2, err2 = quad(LowLevelCallable.from_cython(mod, "_F_integrand3", capsule), 0, h,             # <<<<<<<<<<<<<<
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_v_quad, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 172, __pyx_L5_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_4 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_4)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_2), 2) < 0) __PYX_ERR(0, 172, __pyx_L5_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L12_unpacking_done;
      __pyx_L11_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 172, __pyx_L5_error)
      __pyx_L12_unpacking_done:;
    }
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_res2 = __pyx_t_8;
    __pyx_v_err2 = __pyx_t_9;

    /* "scipy/special/_ellip_harm_2.pyx":175
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 *         res3, err3 = quad(LowLevelCallable.from_cython(mod, "_F_integrand4", capsule), 0, h,             # <<<<<<<<<<<<<<
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_LowLevelCallable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_from_cython); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_mod, __pyx_n_u_F_integrand4, __pyx_v_capsule};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 175, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_mod, __pyx_n_u_F_integrand4, __pyx_v_capsule};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 175, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_mod);
      __Pyx_GIVEREF(__pyx_v_mod);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_mod);
      __Pyx_INCREF(__pyx_n_u_F_integrand4);
      __Pyx_GIVEREF(__pyx_n_u_F_integrand4);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_n_u_F_integrand4);
      __Pyx_INCREF(__pyx_v_capsule);
      __Pyx_GIVEREF(__pyx_v_capsule);
      PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_5, __pyx_v_capsule);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 175, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;

    /* "scipy/special/_ellip_harm_2.pyx":176
 * 
 *         res3, err3 = quad(LowLevelCallable.from_cython(mod, "_F_integrand4", capsule), 0, h,
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)             # <<<<<<<<<<<<<<
 * 
 *     finally:
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_epsabs, __pyx_float_1eneg_300) < 0) __PYX_ERR(0, 176, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_epsrel, __pyx_float_1eneg_15) < 0) __PYX_ERR(0, 176, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_weight, __pyx_n_u_alg) < 0) __PYX_ERR(0, 176, __pyx_L5_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_wvar, __pyx_v_wvar) < 0) __PYX_ERR(0, 176, __pyx_L5_error)

    /* "scipy/special/_ellip_harm_2.pyx":175
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 *         res3, err3 = quad(LowLevelCallable.from_cython(mod, "_F_integrand4", capsule), 0, h,             # <<<<<<<<<<<<<<
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_v_quad, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 175, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 175, __pyx_L5_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L13_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_2 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L13_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 2) < 0) __PYX_ERR(0, 175, __pyx_L5_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L14_unpacking_done;
      __pyx_L13_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 175, __pyx_L5_error)
      __pyx_L14_unpacking_done:;
    }
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L5_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_res3 = __pyx_t_9;
    __pyx_v_err3 = __pyx_t_8;
  }

  /* "scipy/special/_ellip_harm_2.pyx":179
 * 
 *     finally:
 *         free(bufferp)             # <<<<<<<<<<<<<<
 * 
 *     error = 8*(res2*err1 + err2*res1 + res*err3 + res3*err)
 */
  /*finally:*/ {
    /*normal exit:*/{
      free(__pyx_v_bufferp);
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_5 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
      {
        free(__pyx_v_bufferp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "scipy/special/_ellip_harm_2.pyx":181
 *         free(bufferp)
 * 
 *     error = 8*(res2*err1 + err2*res1 + res*err3 + res3*err)             # <<<<<<<<<<<<<<
 *     result = 8*(res1*res2 - res*res3)
 * 
 */
  __pyx_v_error = (8.0 * ((((__pyx_v_res2 * __pyx_v_err1) + (__pyx_v_err2 * __pyx_v_res1)) + (__pyx_v_res * __pyx_v_err3)) + (__pyx_v_res3 * __pyx_v_err)));

  /* "scipy/special/_ellip_harm_2.pyx":182
 * 
 *     error = 8*(res2*err1 + err2*res1 + res*err3 + res3*err)
 *     result = 8*(res1*res2 - res*res3)             # <<<<<<<<<<<<<<
 * 
 *     if error > 10e-8*fabs(result):
 */
  __pyx_v_result = (8.0 * ((__pyx_v_res1 * __pyx_v_res2) - (__pyx_v_res * __pyx_v_res3)));

  /* "scipy/special/_ellip_harm_2.pyx":184
 *     result = 8*(res1*res2 - res*res3)
 * 
 *     if error > 10e-8*fabs(result):             # <<<<<<<<<<<<<<
 *         return nan
 *     return result
 */
  __pyx_t_3 = ((__pyx_v_error > (10e-8 * fabs(__pyx_v_result))) != 0);
  if (__pyx_t_3) {

    /* "scipy/special/_ellip_harm_2.pyx":185
 * 
 *     if error > 10e-8*fabs(result):
 *         return nan             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_nan); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "scipy/special/_ellip_harm_2.pyx":184
 *     result = 8*(res1*res2 - res*res3)
 * 
 *     if error > 10e-8*fabs(result):             # <<<<<<<<<<<<<<
 *         return nan
 *     return result
 */
  }

  /* "scipy/special/_ellip_harm_2.pyx":186
 *     if error > 10e-8*fabs(result):
 *         return nan
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_result); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":137
 * 
 * 
 * def _ellipsoid_norm(double h2, double k2, int n, int p):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("scipy.special._ellip_harm_2._ellipsoid_norm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mod);
  __Pyx_XDECREF(__pyx_v_quad);
  __Pyx_XDECREF(__pyx_v_capsule);
  __Pyx_XDECREF(__pyx_v_wvar);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "scipy/special/_ellip_harm_2.pyx":199
 *     int PyUFunc_getfperr() nogil
 * 
 * cdef public int wrap_PyUFunc_getfperr() nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Call PyUFunc_getfperr in a context where PyUFunc_API array is initialized;
 */

int wrap_PyUFunc_getfperr(void) {
  int __pyx_r;

  /* "scipy/special/_ellip_harm_2.pyx":204
 *     this avoids messing with the UNIQUE_SYMBOL #defines
 *     """
 *     return PyUFunc_getfperr()             # <<<<<<<<<<<<<<
 */
  __pyx_r = PyUFunc_getfperr();
  goto __pyx_L0;

  /* "scipy/special/_ellip_harm_2.pyx":199
 *     int PyUFunc_getfperr() nogil
 * 
 * cdef public int wrap_PyUFunc_getfperr() nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Call PyUFunc_getfperr in a context where PyUFunc_API array is initialized;
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "sf_error.pxd":28
 * 
 * 
 * cdef inline int _sf_error_test_function(int code) nogil:             # <<<<<<<<<<<<<<
 *     """Function that can raise every sf_error category for testing
 *     purposes.
 */

static CYTHON_INLINE int __pyx_f_5scipy_7special_8sf_error__sf_error_test_function(int __pyx_v_code) {
  sf_error_t __pyx_v_sf_error_code;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "sf_error.pxd":35
 *     cdef sf_error_t sf_error_code
 * 
 *     if code < 0 or code >= 10:             # <<<<<<<<<<<<<<
 *         sf_error_code = OTHER
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_code < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_code >= 10) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "sf_error.pxd":36
 * 
 *     if code < 0 or code >= 10:
 *         sf_error_code = OTHER             # <<<<<<<<<<<<<<
 *     else:
 *         sf_error_code = <sf_error_t>code
 */
    __pyx_v_sf_error_code = SF_ERROR_OTHER;

    /* "sf_error.pxd":35
 *     cdef sf_error_t sf_error_code
 * 
 *     if code < 0 or code >= 10:             # <<<<<<<<<<<<<<
 *         sf_error_code = OTHER
 *     else:
 */
    goto __pyx_L3;
  }

  /* "sf_error.pxd":38
 *         sf_error_code = OTHER
 *     else:
 *         sf_error_code = <sf_error_t>code             # <<<<<<<<<<<<<<
 *     error('_err_test_function', sf_error_code, NULL)
 *     return 0
 */
  /*else*/ {
    __pyx_v_sf_error_code = ((sf_error_t)__pyx_v_code);
  }
  __pyx_L3:;

  /* "sf_error.pxd":39
 *     else:
 *         sf_error_code = <sf_error_t>code
 *     error('_err_test_function', sf_error_code, NULL)             # <<<<<<<<<<<<<<
 *     return 0
 */
  sf_error(((char *)"_err_test_function"), __pyx_v_sf_error_code, NULL);

  /* "sf_error.pxd":40
 *         sf_error_code = <sf_error_t>code
 *     error('_err_test_function', sf_error_code, NULL)
 *     return 0             # <<<<<<<<<<<<<<
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "sf_error.pxd":28
 * 
 * 
 * cdef inline int _sf_error_test_function(int code) nogil:             # <<<<<<<<<<<<<<
 *     """Function that can raise every sf_error category for testing
 *     purposes.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "_ellip_harm.pxd":51
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cdef inline double* lame_coefficients(double h2, double k2, int n, int p,             # <<<<<<<<<<<<<<
 *                                       void **bufferp, double signm,
 *                                       double signn) nogil:
 */

static CYTHON_INLINE double *__pyx_f_5scipy_7special_11_ellip_harm_lame_coefficients(double __pyx_v_h2, double __pyx_v_k2, int __pyx_v_n, int __pyx_v_p, void **__pyx_v_bufferp, double __pyx_v_signm, double __pyx_v_signn) {
  double __pyx_v_alpha;
  double __pyx_v_beta;
  double __pyx_v_gamma;
  double __pyx_v_tol;
  double __pyx_v_vl;
  double __pyx_v_vu;
  CBLAS_INT __pyx_v_r;
  CBLAS_INT __pyx_v_tp;
  CBLAS_INT __pyx_v_j;
  CBLAS_INT __pyx_v_size;
  CBLAS_INT __pyx_v_i;
  CBLAS_INT __pyx_v_info;
  CBLAS_INT __pyx_v_lwork;
  CBLAS_INT __pyx_v_liwork;
  CBLAS_INT __pyx_v_c;
  Py_UNICODE __pyx_v_t;
  void *__pyx_v_buffer;
  double *__pyx_v_g;
  double *__pyx_v_d;
  double *__pyx_v_f;
  double *__pyx_v_ss;
  double *__pyx_v_w;
  double *__pyx_v_dd;
  double *__pyx_v_eigv;
  double *__pyx_v_work;
  CBLAS_INT *__pyx_v_iwork;
  CBLAS_INT *__pyx_v_isuppz;
  double *__pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  Py_UNICODE __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;
  CBLAS_INT __pyx_t_6;
  long __pyx_t_7;
  CBLAS_INT __pyx_t_8;
  CBLAS_INT __pyx_t_9;
  CBLAS_INT __pyx_t_10;

  /* "_ellip_harm.pxd":57
 *     # Ensure that the caller can safely call free(*bufferp) even if an
 *     # invalid argument is found in the following validation code.
 *     bufferp[0] = NULL             # <<<<<<<<<<<<<<
 * 
 *     if n < 0:
 */
  (__pyx_v_bufferp[0]) = NULL;

  /* "_ellip_harm.pxd":59
 *     bufferp[0] = NULL
 * 
 *     if n < 0:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for n")
 *         return NULL
 */
  __pyx_t_1 = ((__pyx_v_n < 0) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":60
 * 
 *     if n < 0:
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for n")             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */
    sf_error(((char *)"ellip_harm"), SF_ERROR_ARG, ((char *)"invalid value for n"));

    /* "_ellip_harm.pxd":61
 *     if n < 0:
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for n")
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     if p < 1 or p > 2*n + 1:
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "_ellip_harm.pxd":59
 *     bufferp[0] = NULL
 * 
 *     if n < 0:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for n")
 *         return NULL
 */
  }

  /* "_ellip_harm.pxd":63
 *         return NULL
 * 
 *     if p < 1 or p > 2*n + 1:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for p")
 *         return NULL
 */
  __pyx_t_2 = ((__pyx_v_p < 1) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_p > ((2 * __pyx_v_n) + 1)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":64
 * 
 *     if p < 1 or p > 2*n + 1:
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for p")             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */
    sf_error(((char *)"ellip_harm"), SF_ERROR_ARG, ((char *)"invalid value for p"));

    /* "_ellip_harm.pxd":65
 *     if p < 1 or p > 2*n + 1:
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for p")
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     if fabs(signm) != 1 or fabs(signn) != 1:
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "_ellip_harm.pxd":63
 *         return NULL
 * 
 *     if p < 1 or p > 2*n + 1:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid value for p")
 *         return NULL
 */
  }

  /* "_ellip_harm.pxd":67
 *         return NULL
 * 
 *     if fabs(signm) != 1 or fabs(signn) != 1:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid signm or signn")
 *         return NULL
 */
  __pyx_t_2 = ((fabs(__pyx_v_signm) != 1.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_2 = ((fabs(__pyx_v_signn) != 1.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":68
 * 
 *     if fabs(signm) != 1 or fabs(signn) != 1:
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid signm or signn")             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */
    sf_error(((char *)"ellip_harm"), SF_ERROR_ARG, ((char *)"invalid signm or signn"));

    /* "_ellip_harm.pxd":69
 *     if fabs(signm) != 1 or fabs(signn) != 1:
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid signm or signn")
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef double s2, alpha, beta, gamma, lamba_romain, pp, psi, t1, tol, vl, vu
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "_ellip_harm.pxd":67
 *         return NULL
 * 
 *     if fabs(signm) != 1 or fabs(signn) != 1:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.ARG, "invalid signm or signn")
 *         return NULL
 */
  }

  /* "_ellip_harm.pxd":75
 *     cdef Py_UNICODE t
 * 
 *     r = n/2             # <<<<<<<<<<<<<<
 *     alpha = h2
 *     beta = k2 - h2
 */
  __pyx_v_r = (((long)__pyx_v_n) / 2);

  /* "_ellip_harm.pxd":76
 * 
 *     r = n/2
 *     alpha = h2             # <<<<<<<<<<<<<<
 *     beta = k2 - h2
 *     gamma = alpha - beta
 */
  __pyx_v_alpha = __pyx_v_h2;

  /* "_ellip_harm.pxd":77
 *     r = n/2
 *     alpha = h2
 *     beta = k2 - h2             # <<<<<<<<<<<<<<
 *     gamma = alpha - beta
 * 
 */
  __pyx_v_beta = (__pyx_v_k2 - __pyx_v_h2);

  /* "_ellip_harm.pxd":78
 *     alpha = h2
 *     beta = k2 - h2
 *     gamma = alpha - beta             # <<<<<<<<<<<<<<
 * 
 *     if p - 1 < r + 1:
 */
  __pyx_v_gamma = (__pyx_v_alpha - __pyx_v_beta);

  /* "_ellip_harm.pxd":80
 *     gamma = alpha - beta
 * 
 *     if p - 1 < r + 1:             # <<<<<<<<<<<<<<
 *         t, tp, size = 'K', p, r + 1
 *     elif p - 1 < (n - r) + (r + 1):
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < (__pyx_v_r + 1)) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":81
 * 
 *     if p - 1 < r + 1:
 *         t, tp, size = 'K', p, r + 1             # <<<<<<<<<<<<<<
 *     elif p - 1 < (n - r) + (r + 1):
 *         t, tp, size = 'L', p - (r + 1), n - r
 */
    __pyx_t_3 = 75;
    __pyx_t_4 = __pyx_v_p;
    __pyx_t_5 = (__pyx_v_r + 1);
    __pyx_v_t = __pyx_t_3;
    __pyx_v_tp = __pyx_t_4;
    __pyx_v_size = __pyx_t_5;

    /* "_ellip_harm.pxd":80
 *     gamma = alpha - beta
 * 
 *     if p - 1 < r + 1:             # <<<<<<<<<<<<<<
 *         t, tp, size = 'K', p, r + 1
 *     elif p - 1 < (n - r) + (r + 1):
 */
    goto __pyx_L10;
  }

  /* "_ellip_harm.pxd":82
 *     if p - 1 < r + 1:
 *         t, tp, size = 'K', p, r + 1
 *     elif p - 1 < (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         t, tp, size = 'L', p - (r + 1), n - r
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < ((__pyx_v_n - __pyx_v_r) + (__pyx_v_r + 1))) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":83
 *         t, tp, size = 'K', p, r + 1
 *     elif p - 1 < (n - r) + (r + 1):
 *         t, tp, size = 'L', p - (r + 1), n - r             # <<<<<<<<<<<<<<
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         t, tp, size = 'M', p - (n - r) - (r + 1), n - r
 */
    __pyx_t_3 = 76;
    __pyx_t_5 = (__pyx_v_p - (__pyx_v_r + 1));
    __pyx_t_6 = (__pyx_v_n - __pyx_v_r);
    __pyx_v_t = __pyx_t_3;
    __pyx_v_tp = __pyx_t_5;
    __pyx_v_size = __pyx_t_6;

    /* "_ellip_harm.pxd":82
 *     if p - 1 < r + 1:
 *         t, tp, size = 'K', p, r + 1
 *     elif p - 1 < (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         t, tp, size = 'L', p - (r + 1), n - r
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 */
    goto __pyx_L10;
  }

  /* "_ellip_harm.pxd":84
 *     elif p - 1 < (n - r) + (r + 1):
 *         t, tp, size = 'L', p - (r + 1), n - r
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         t, tp, size = 'M', p - (n - r) - (r + 1), n - r
 *     elif p - 1 < 2*n + 1:
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < (((__pyx_v_n - __pyx_v_r) + (__pyx_v_n - __pyx_v_r)) + (__pyx_v_r + 1))) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":85
 *         t, tp, size = 'L', p - (r + 1), n - r
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         t, tp, size = 'M', p - (n - r) - (r + 1), n - r             # <<<<<<<<<<<<<<
 *     elif p - 1 < 2*n + 1:
 *         t, tp, size = 'N', p - (n - r) - (n - r) - (r + 1), r
 */
    __pyx_t_3 = 77;
    __pyx_t_5 = ((__pyx_v_p - (__pyx_v_n - __pyx_v_r)) - (__pyx_v_r + 1));
    __pyx_t_6 = (__pyx_v_n - __pyx_v_r);
    __pyx_v_t = __pyx_t_3;
    __pyx_v_tp = __pyx_t_5;
    __pyx_v_size = __pyx_t_6;

    /* "_ellip_harm.pxd":84
 *     elif p - 1 < (n - r) + (r + 1):
 *         t, tp, size = 'L', p - (r + 1), n - r
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         t, tp, size = 'M', p - (n - r) - (r + 1), n - r
 *     elif p - 1 < 2*n + 1:
 */
    goto __pyx_L10;
  }

  /* "_ellip_harm.pxd":86
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         t, tp, size = 'M', p - (n - r) - (r + 1), n - r
 *     elif p - 1 < 2*n + 1:             # <<<<<<<<<<<<<<
 *         t, tp, size = 'N', p - (n - r) - (n - r) - (r + 1), r
 * 
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < ((2 * __pyx_v_n) + 1)) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":87
 *         t, tp, size = 'M', p - (n - r) - (r + 1), n - r
 *     elif p - 1 < 2*n + 1:
 *         t, tp, size = 'N', p - (n - r) - (n - r) - (r + 1), r             # <<<<<<<<<<<<<<
 * 
 *     lwork = 60*size
 */
    __pyx_t_3 = 78;
    __pyx_t_5 = (((__pyx_v_p - (__pyx_v_n - __pyx_v_r)) - (__pyx_v_n - __pyx_v_r)) - (__pyx_v_r + 1));
    __pyx_t_6 = __pyx_v_r;
    __pyx_v_t = __pyx_t_3;
    __pyx_v_tp = __pyx_t_5;
    __pyx_v_size = __pyx_t_6;

    /* "_ellip_harm.pxd":86
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         t, tp, size = 'M', p - (n - r) - (r + 1), n - r
 *     elif p - 1 < 2*n + 1:             # <<<<<<<<<<<<<<
 *         t, tp, size = 'N', p - (n - r) - (n - r) - (r + 1), r
 * 
 */
  }
  __pyx_L10:;

  /* "_ellip_harm.pxd":89
 *         t, tp, size = 'N', p - (n - r) - (n - r) - (r + 1), r
 * 
 *     lwork = 60*size             # <<<<<<<<<<<<<<
 *     liwork = 30*size
 *     tol = 0.0
 */
  __pyx_v_lwork = (60 * __pyx_v_size);

  /* "_ellip_harm.pxd":90
 * 
 *     lwork = 60*size
 *     liwork = 30*size             # <<<<<<<<<<<<<<
 *     tol = 0.0
 *     vl = 0
 */
  __pyx_v_liwork = (30 * __pyx_v_size);

  /* "_ellip_harm.pxd":91
 *     lwork = 60*size
 *     liwork = 30*size
 *     tol = 0.0             # <<<<<<<<<<<<<<
 *     vl = 0
 *     vu = 0
 */
  __pyx_v_tol = 0.0;

  /* "_ellip_harm.pxd":92
 *     liwork = 30*size
 *     tol = 0.0
 *     vl = 0             # <<<<<<<<<<<<<<
 *     vu = 0
 * 
 */
  __pyx_v_vl = 0.0;

  /* "_ellip_harm.pxd":93
 *     tol = 0.0
 *     vl = 0
 *     vu = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void *buffer = malloc((sizeof(double)*(7*size + lwork))
 */
  __pyx_v_vu = 0.0;

  /* "_ellip_harm.pxd":95
 *     vu = 0
 * 
 *     cdef void *buffer = malloc((sizeof(double)*(7*size + lwork))             # <<<<<<<<<<<<<<
 *                                + (sizeof(CBLAS_INT)*(2*size + liwork)))
 *     bufferp[0] = buffer
 */
  __pyx_v_buffer = malloc((((sizeof(double)) * ((7 * __pyx_v_size) + __pyx_v_lwork)) + ((sizeof(CBLAS_INT)) * ((2 * __pyx_v_size) + __pyx_v_liwork))));

  /* "_ellip_harm.pxd":97
 *     cdef void *buffer = malloc((sizeof(double)*(7*size + lwork))
 *                                + (sizeof(CBLAS_INT)*(2*size + liwork)))
 *     bufferp[0] = buffer             # <<<<<<<<<<<<<<
 *     if not buffer:
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")
 */
  (__pyx_v_bufferp[0]) = __pyx_v_buffer;

  /* "_ellip_harm.pxd":98
 *                                + (sizeof(CBLAS_INT)*(2*size + liwork)))
 *     bufferp[0] = buffer
 *     if not buffer:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")
 *         return NULL
 */
  __pyx_t_1 = ((!(__pyx_v_buffer != 0)) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":99
 *     bufferp[0] = buffer
 *     if not buffer:
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */
    sf_error(((char *)"ellip_harm"), SF_ERROR_NO_RESULT, ((char *)"failed to allocate memory"));

    /* "_ellip_harm.pxd":100
 *     if not buffer:
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef double *g = <double *>buffer
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "_ellip_harm.pxd":98
 *                                + (sizeof(CBLAS_INT)*(2*size + liwork)))
 *     bufferp[0] = buffer
 *     if not buffer:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")
 *         return NULL
 */
  }

  /* "_ellip_harm.pxd":102
 *         return NULL
 * 
 *     cdef double *g = <double *>buffer             # <<<<<<<<<<<<<<
 *     cdef double *d = g + size
 *     cdef double *f = d + size
 */
  __pyx_v_g = ((double *)__pyx_v_buffer);

  /* "_ellip_harm.pxd":103
 * 
 *     cdef double *g = <double *>buffer
 *     cdef double *d = g + size             # <<<<<<<<<<<<<<
 *     cdef double *f = d + size
 *     cdef double *ss =  f + size
 */
  __pyx_v_d = (__pyx_v_g + __pyx_v_size);

  /* "_ellip_harm.pxd":104
 *     cdef double *g = <double *>buffer
 *     cdef double *d = g + size
 *     cdef double *f = d + size             # <<<<<<<<<<<<<<
 *     cdef double *ss =  f + size
 *     cdef double *w =  ss + size
 */
  __pyx_v_f = (__pyx_v_d + __pyx_v_size);

  /* "_ellip_harm.pxd":105
 *     cdef double *d = g + size
 *     cdef double *f = d + size
 *     cdef double *ss =  f + size             # <<<<<<<<<<<<<<
 *     cdef double *w =  ss + size
 *     cdef double *dd = w + size
 */
  __pyx_v_ss = (__pyx_v_f + __pyx_v_size);

  /* "_ellip_harm.pxd":106
 *     cdef double *f = d + size
 *     cdef double *ss =  f + size
 *     cdef double *w =  ss + size             # <<<<<<<<<<<<<<
 *     cdef double *dd = w + size
 *     cdef double *eigv = dd + size
 */
  __pyx_v_w = (__pyx_v_ss + __pyx_v_size);

  /* "_ellip_harm.pxd":107
 *     cdef double *ss =  f + size
 *     cdef double *w =  ss + size
 *     cdef double *dd = w + size             # <<<<<<<<<<<<<<
 *     cdef double *eigv = dd + size
 *     cdef double *work = eigv + size
 */
  __pyx_v_dd = (__pyx_v_w + __pyx_v_size);

  /* "_ellip_harm.pxd":108
 *     cdef double *w =  ss + size
 *     cdef double *dd = w + size
 *     cdef double *eigv = dd + size             # <<<<<<<<<<<<<<
 *     cdef double *work = eigv + size
 * 
 */
  __pyx_v_eigv = (__pyx_v_dd + __pyx_v_size);

  /* "_ellip_harm.pxd":109
 *     cdef double *dd = w + size
 *     cdef double *eigv = dd + size
 *     cdef double *work = eigv + size             # <<<<<<<<<<<<<<
 * 
 *     cdef CBLAS_INT *iwork = <CBLAS_INT *>(work + lwork)
 */
  __pyx_v_work = (__pyx_v_eigv + __pyx_v_size);

  /* "_ellip_harm.pxd":111
 *     cdef double *work = eigv + size
 * 
 *     cdef CBLAS_INT *iwork = <CBLAS_INT *>(work + lwork)             # <<<<<<<<<<<<<<
 *     cdef CBLAS_INT *isuppz = iwork + liwork
 * 
 */
  __pyx_v_iwork = ((CBLAS_INT *)(__pyx_v_work + __pyx_v_lwork));

  /* "_ellip_harm.pxd":112
 * 
 *     cdef CBLAS_INT *iwork = <CBLAS_INT *>(work + lwork)
 *     cdef CBLAS_INT *isuppz = iwork + liwork             # <<<<<<<<<<<<<<
 * 
 *     if t == 'K':
 */
  __pyx_v_isuppz = (__pyx_v_iwork + __pyx_v_liwork);

  /* "_ellip_harm.pxd":114
 *     cdef CBLAS_INT *isuppz = iwork + liwork
 * 
 *     if t == 'K':             # <<<<<<<<<<<<<<
 *         for j in range(0, r + 1):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 */
  switch (__pyx_v_t) {
    case 75:

    /* "_ellip_harm.pxd":115
 * 
 *     if t == 'K':
 *         for j in range(0, r + 1):             # <<<<<<<<<<<<<<
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:
 */
    __pyx_t_5 = (__pyx_v_r + 1);
    __pyx_t_7 = __pyx_t_5;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_7; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "_ellip_harm.pxd":116
 *     if t == 'K':
 *         for j in range(0, r + 1):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)             # <<<<<<<<<<<<<<
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 */
      (__pyx_v_g[__pyx_v_j]) = (((-((2 * __pyx_v_j) + 2)) * ((2 * __pyx_v_j) + 1)) * __pyx_v_beta);

      /* "_ellip_harm.pxd":117
 *         for j in range(0, r + 1):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = ((2*r + 1)*(2*r + 2) - 4*j*j)*alpha + (2*j + 1)*(2*j + 1)*beta
 */
      __pyx_t_1 = ((__pyx_v_n % 2) != 0);
      if (__pyx_t_1) {

        /* "_ellip_harm.pxd":118
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))             # <<<<<<<<<<<<<<
 *                d[j] = ((2*r + 1)*(2*r + 2) - 4*j*j)*alpha + (2*j + 1)*(2*j + 1)*beta
 *            else:
 */
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * ((2 * (__pyx_v_r - (__pyx_v_j + 1))) + 2)) * ((2 * ((__pyx_v_j + 1) + __pyx_v_r)) + 1));

        /* "_ellip_harm.pxd":119
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = ((2*r + 1)*(2*r + 2) - 4*j*j)*alpha + (2*j + 1)*(2*j + 1)*beta             # <<<<<<<<<<<<<<
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*(r + (j + 1)) - 1))
 */
        (__pyx_v_d[__pyx_v_j]) = ((((((2 * __pyx_v_r) + 1) * ((2 * __pyx_v_r) + 2)) - ((4 * __pyx_v_j) * __pyx_v_j)) * __pyx_v_alpha) + ((((2 * __pyx_v_j) + 1) * ((2 * __pyx_v_j) + 1)) * __pyx_v_beta));

        /* "_ellip_harm.pxd":117
 *         for j in range(0, r + 1):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = ((2*r + 1)*(2*r + 2) - 4*j*j)*alpha + (2*j + 1)*(2*j + 1)*beta
 */
        goto __pyx_L14;
      }

      /* "_ellip_harm.pxd":121
 *                d[j] = ((2*r + 1)*(2*r + 2) - 4*j*j)*alpha + (2*j + 1)*(2*j + 1)*beta
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*(r + (j + 1)) - 1))             # <<<<<<<<<<<<<<
 *                d[j] = 2*r*(2*r + 1)*alpha - 4*j*j*gamma
 * 
 */
      /*else*/ {
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * ((2 * (__pyx_v_r - (__pyx_v_j + 1))) + 2)) * ((2 * (__pyx_v_r + (__pyx_v_j + 1))) - 1));

        /* "_ellip_harm.pxd":122
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*(r + (j + 1)) - 1))
 *                d[j] = 2*r*(2*r + 1)*alpha - 4*j*j*gamma             # <<<<<<<<<<<<<<
 * 
 *     elif t == 'L':
 */
        (__pyx_v_d[__pyx_v_j]) = ((((2 * __pyx_v_r) * ((2 * __pyx_v_r) + 1)) * __pyx_v_alpha) - (((4 * __pyx_v_j) * __pyx_v_j) * __pyx_v_gamma));
      }
      __pyx_L14:;
    }

    /* "_ellip_harm.pxd":114
 *     cdef CBLAS_INT *isuppz = iwork + liwork
 * 
 *     if t == 'K':             # <<<<<<<<<<<<<<
 *         for j in range(0, r + 1):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 */
    break;
    case 76:

    /* "_ellip_harm.pxd":125
 * 
 *     elif t == 'L':
 *         for j in range(0, n - r):             # <<<<<<<<<<<<<<
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:
 */
    __pyx_t_6 = (__pyx_v_n - __pyx_v_r);
    __pyx_t_8 = __pyx_t_6;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_j = __pyx_t_9;

      /* "_ellip_harm.pxd":126
 *     elif t == 'L':
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)             # <<<<<<<<<<<<<<
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 */
      (__pyx_v_g[__pyx_v_j]) = (((-((2 * __pyx_v_j) + 2)) * ((2 * __pyx_v_j) + 3)) * __pyx_v_beta);

      /* "_ellip_harm.pxd":127
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 1)*(2*j + 1)*gamma
 */
      __pyx_t_1 = ((__pyx_v_n % 2) != 0);
      if (__pyx_t_1) {

        /* "_ellip_harm.pxd":128
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))             # <<<<<<<<<<<<<<
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 1)*(2*j + 1)*gamma
 *            else:
 */
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * ((2 * (__pyx_v_r - (__pyx_v_j + 1))) + 2)) * ((2 * ((__pyx_v_j + 1) + __pyx_v_r)) + 1));

        /* "_ellip_harm.pxd":129
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 1)*(2*j + 1)*gamma             # <<<<<<<<<<<<<<
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))
 */
        (__pyx_v_d[__pyx_v_j]) = (((((2 * __pyx_v_r) + 1) * ((2 * __pyx_v_r) + 2)) * __pyx_v_alpha) - ((((2 * __pyx_v_j) + 1) * ((2 * __pyx_v_j) + 1)) * __pyx_v_gamma));

        /* "_ellip_harm.pxd":127
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r- (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 1)*(2*j + 1)*gamma
 */
        goto __pyx_L17;
      }

      /* "_ellip_harm.pxd":131
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 1)*(2*j + 1)*gamma
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))             # <<<<<<<<<<<<<<
 *                d[j] = (2*r*(2*r + 1) - (2*j + 1)*(2*j + 1))*alpha + (2*j + 2)*(2*j + 2)*beta
 * 
 */
      /*else*/ {
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * (2 * (__pyx_v_r - (__pyx_v_j + 1)))) * ((2 * (__pyx_v_r + (__pyx_v_j + 1))) + 1));

        /* "_ellip_harm.pxd":132
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))
 *                d[j] = (2*r*(2*r + 1) - (2*j + 1)*(2*j + 1))*alpha + (2*j + 2)*(2*j + 2)*beta             # <<<<<<<<<<<<<<
 * 
 *     elif t == 'M':
 */
        (__pyx_v_d[__pyx_v_j]) = (((((2 * __pyx_v_r) * ((2 * __pyx_v_r) + 1)) - (((2 * __pyx_v_j) + 1) * ((2 * __pyx_v_j) + 1))) * __pyx_v_alpha) + ((((2 * __pyx_v_j) + 2) * ((2 * __pyx_v_j) + 2)) * __pyx_v_beta));
      }
      __pyx_L17:;
    }

    /* "_ellip_harm.pxd":124
 *                d[j] = 2*r*(2*r + 1)*alpha - 4*j*j*gamma
 * 
 *     elif t == 'L':             # <<<<<<<<<<<<<<
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 */
    break;
    case 77:

    /* "_ellip_harm.pxd":135
 * 
 *     elif t == 'M':
 *         for j in range(0, n - r):             # <<<<<<<<<<<<<<
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:
 */
    __pyx_t_6 = (__pyx_v_n - __pyx_v_r);
    __pyx_t_8 = __pyx_t_6;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_j = __pyx_t_9;

      /* "_ellip_harm.pxd":136
 *     elif t == 'M':
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)             # <<<<<<<<<<<<<<
 *            if n%2:
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 */
      (__pyx_v_g[__pyx_v_j]) = (((-((2 * __pyx_v_j) + 2)) * ((2 * __pyx_v_j) + 1)) * __pyx_v_beta);

      /* "_ellip_harm.pxd":137
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = ((2*r + 1)*(2*r + 2) - (2*j + 1)*(2*j + 1))*alpha + 4*j*j*beta
 */
      __pyx_t_1 = ((__pyx_v_n % 2) != 0);
      if (__pyx_t_1) {

        /* "_ellip_harm.pxd":138
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*((j + 1) + r) + 1))             # <<<<<<<<<<<<<<
 *                d[j] = ((2*r + 1)*(2*r + 2) - (2*j + 1)*(2*j + 1))*alpha + 4*j*j*beta
 *            else:
 */
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * ((2 * (__pyx_v_r - (__pyx_v_j + 1))) + 2)) * ((2 * ((__pyx_v_j + 1) + __pyx_v_r)) + 1));

        /* "_ellip_harm.pxd":139
 *            if n%2:
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = ((2*r + 1)*(2*r + 2) - (2*j + 1)*(2*j + 1))*alpha + 4*j*j*beta             # <<<<<<<<<<<<<<
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))
 */
        (__pyx_v_d[__pyx_v_j]) = ((((((2 * __pyx_v_r) + 1) * ((2 * __pyx_v_r) + 2)) - (((2 * __pyx_v_j) + 1) * ((2 * __pyx_v_j) + 1))) * __pyx_v_alpha) + (((4 * __pyx_v_j) * __pyx_v_j) * __pyx_v_beta));

        /* "_ellip_harm.pxd":137
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r - (j + 1)) + 2)*(2*((j + 1) + r) + 1))
 *                d[j] = ((2*r + 1)*(2*r + 2) - (2*j + 1)*(2*j + 1))*alpha + 4*j*j*beta
 */
        goto __pyx_L20;
      }

      /* "_ellip_harm.pxd":141
 *                d[j] = ((2*r + 1)*(2*r + 2) - (2*j + 1)*(2*j + 1))*alpha + 4*j*j*beta
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))             # <<<<<<<<<<<<<<
 *                d[j] = 2*r*(2*r + 1)*alpha - (2*j + 1)*(2*j + 1)*gamma
 * 
 */
      /*else*/ {
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * (2 * (__pyx_v_r - (__pyx_v_j + 1)))) * ((2 * (__pyx_v_r + (__pyx_v_j + 1))) + 1));

        /* "_ellip_harm.pxd":142
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))
 *                d[j] = 2*r*(2*r + 1)*alpha - (2*j + 1)*(2*j + 1)*gamma             # <<<<<<<<<<<<<<
 * 
 *     elif t == 'N':
 */
        (__pyx_v_d[__pyx_v_j]) = ((((2 * __pyx_v_r) * ((2 * __pyx_v_r) + 1)) * __pyx_v_alpha) - ((((2 * __pyx_v_j) + 1) * ((2 * __pyx_v_j) + 1)) * __pyx_v_gamma));
      }
      __pyx_L20:;
    }

    /* "_ellip_harm.pxd":134
 *                d[j] = (2*r*(2*r + 1) - (2*j + 1)*(2*j + 1))*alpha + (2*j + 2)*(2*j + 2)*beta
 * 
 *     elif t == 'M':             # <<<<<<<<<<<<<<
 *         for j in range(0, n - r):
 *            g[j] = (-(2*j + 2)*(2*j + 1)*beta)
 */
    break;
    case 78:

    /* "_ellip_harm.pxd":145
 * 
 *     elif t == 'N':
 *         for j in range(0, r):             # <<<<<<<<<<<<<<
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:
 */
    __pyx_t_6 = __pyx_v_r;
    __pyx_t_8 = __pyx_t_6;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_j = __pyx_t_9;

      /* "_ellip_harm.pxd":146
 *     elif t == 'N':
 *         for j in range(0, r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)             # <<<<<<<<<<<<<<
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)))*(2*((j + 1) + r) + 3))
 */
      (__pyx_v_g[__pyx_v_j]) = (((-((2 * __pyx_v_j) + 2)) * ((2 * __pyx_v_j) + 3)) * __pyx_v_beta);

      /* "_ellip_harm.pxd":147
 *         for j in range(0, r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r- (j + 1)))*(2*((j + 1) + r) + 3))
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 2)*(2*j + 2)*gamma
 */
      __pyx_t_1 = ((__pyx_v_n % 2) != 0);
      if (__pyx_t_1) {

        /* "_ellip_harm.pxd":148
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)))*(2*((j + 1) + r) + 3))             # <<<<<<<<<<<<<<
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 2)*(2*j + 2)*gamma
 *            else:
 */
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * (2 * (__pyx_v_r - (__pyx_v_j + 1)))) * ((2 * ((__pyx_v_j + 1) + __pyx_v_r)) + 3));

        /* "_ellip_harm.pxd":149
 *            if n%2:
 *                f[j] = (-alpha*(2*(r- (j + 1)))*(2*((j + 1) + r) + 3))
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 2)*(2*j + 2)*gamma             # <<<<<<<<<<<<<<
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))
 */
        (__pyx_v_d[__pyx_v_j]) = (((((2 * __pyx_v_r) + 1) * ((2 * __pyx_v_r) + 2)) * __pyx_v_alpha) - ((((2 * __pyx_v_j) + 2) * ((2 * __pyx_v_j) + 2)) * __pyx_v_gamma));

        /* "_ellip_harm.pxd":147
 *         for j in range(0, r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 *            if n%2:             # <<<<<<<<<<<<<<
 *                f[j] = (-alpha*(2*(r- (j + 1)))*(2*((j + 1) + r) + 3))
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 2)*(2*j + 2)*gamma
 */
        goto __pyx_L23;
      }

      /* "_ellip_harm.pxd":151
 *                d[j] = (2*r + 1)*(2*r + 2)*alpha - (2*j + 2)*(2*j + 2)*gamma
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))             # <<<<<<<<<<<<<<
 *                d[j] = 2*r*(2*r + 1)*alpha - (2*j + 2)*(2*j +2)*alpha + (2*j + 1)*(2*j + 1)*beta
 * 
 */
      /*else*/ {
        (__pyx_v_f[__pyx_v_j]) = (((-__pyx_v_alpha) * (2 * (__pyx_v_r - (__pyx_v_j + 1)))) * ((2 * (__pyx_v_r + (__pyx_v_j + 1))) + 1));

        /* "_ellip_harm.pxd":152
 *            else:
 *                f[j] = (-alpha*(2*(r - (j + 1)))*(2*(r+(j + 1)) + 1))
 *                d[j] = 2*r*(2*r + 1)*alpha - (2*j + 2)*(2*j +2)*alpha + (2*j + 1)*(2*j + 1)*beta             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, size):
 */
        (__pyx_v_d[__pyx_v_j]) = (((((2 * __pyx_v_r) * ((2 * __pyx_v_r) + 1)) * __pyx_v_alpha) - ((((2 * __pyx_v_j) + 2) * ((2 * __pyx_v_j) + 2)) * __pyx_v_alpha)) + ((((2 * __pyx_v_j) + 1) * ((2 * __pyx_v_j) + 1)) * __pyx_v_beta));
      }
      __pyx_L23:;
    }

    /* "_ellip_harm.pxd":144
 *                d[j] = 2*r*(2*r + 1)*alpha - (2*j + 1)*(2*j + 1)*gamma
 * 
 *     elif t == 'N':             # <<<<<<<<<<<<<<
 *         for j in range(0, r):
 *            g[j] = (-(2*j + 2)*(2*j + 3)*beta)
 */
    break;
    default: break;
  }

  /* "_ellip_harm.pxd":154
 *                d[j] = 2*r*(2*r + 1)*alpha - (2*j + 2)*(2*j +2)*alpha + (2*j + 1)*(2*j + 1)*beta
 * 
 *     for i in range(0, size):             # <<<<<<<<<<<<<<
 *         if i == 0:
 *            ss[i] = 1
 */
  __pyx_t_6 = __pyx_v_size;
  __pyx_t_8 = __pyx_t_6;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "_ellip_harm.pxd":155
 * 
 *     for i in range(0, size):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *            ss[i] = 1
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_i == 0) != 0);
    if (__pyx_t_1) {

      /* "_ellip_harm.pxd":156
 *     for i in range(0, size):
 *         if i == 0:
 *            ss[i] = 1             # <<<<<<<<<<<<<<
 *         else:
 *            ss[i] = sqrt(g[i - 1]/f[i - 1])*ss[i - 1]
 */
      (__pyx_v_ss[__pyx_v_i]) = 1.0;

      /* "_ellip_harm.pxd":155
 * 
 *     for i in range(0, size):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *            ss[i] = 1
 *         else:
 */
      goto __pyx_L26;
    }

    /* "_ellip_harm.pxd":158
 *            ss[i] = 1
 *         else:
 *            ss[i] = sqrt(g[i - 1]/f[i - 1])*ss[i - 1]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, size-1):
 */
    /*else*/ {
      (__pyx_v_ss[__pyx_v_i]) = (sqrt(((__pyx_v_g[(__pyx_v_i - 1)]) / (__pyx_v_f[(__pyx_v_i - 1)]))) * (__pyx_v_ss[(__pyx_v_i - 1)]));
    }
    __pyx_L26:;
  }

  /* "_ellip_harm.pxd":160
 *            ss[i] = sqrt(g[i - 1]/f[i - 1])*ss[i - 1]
 * 
 *     for i in range(0, size-1):             # <<<<<<<<<<<<<<
 *         dd[i] = g[i]*ss[i]/ss[i+1]
 * 
 */
  __pyx_t_5 = (__pyx_v_size - 1);
  __pyx_t_7 = __pyx_t_5;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_7; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "_ellip_harm.pxd":161
 * 
 *     for i in range(0, size-1):
 *         dd[i] = g[i]*ss[i]/ss[i+1]             # <<<<<<<<<<<<<<
 * 
 *     c_dstevr("V", "I", &size, d, dd, &vl, &vu, &tp, &tp, &tol, &c, w, eigv,
 */
    (__pyx_v_dd[__pyx_v_i]) = (((__pyx_v_g[__pyx_v_i]) * (__pyx_v_ss[__pyx_v_i])) / (__pyx_v_ss[(__pyx_v_i + 1)]));
  }

  /* "_ellip_harm.pxd":163
 *         dd[i] = g[i]*ss[i]/ss[i+1]
 * 
 *     c_dstevr("V", "I", &size, d, dd, &vl, &vu, &tp, &tp, &tol, &c, w, eigv,             # <<<<<<<<<<<<<<
 *              &size, isuppz, work, &lwork, iwork, &liwork, &info)
 * 
 */
  c_dstevr(((char *)"V"), ((char *)"I"), (&__pyx_v_size), __pyx_v_d, __pyx_v_dd, (&__pyx_v_vl), (&__pyx_v_vu), (&__pyx_v_tp), (&__pyx_v_tp), (&__pyx_v_tol), (&__pyx_v_c), __pyx_v_w, __pyx_v_eigv, (&__pyx_v_size), __pyx_v_isuppz, __pyx_v_work, (&__pyx_v_lwork), __pyx_v_iwork, (&__pyx_v_liwork), (&__pyx_v_info));

  /* "_ellip_harm.pxd":166
 *              &size, isuppz, work, &lwork, iwork, &liwork, &info)
 * 
 *     if info != 0:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")
 *         return NULL
 */
  __pyx_t_1 = ((__pyx_v_info != 0) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":167
 * 
 *     if info != 0:
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */
    sf_error(((char *)"ellip_harm"), SF_ERROR_NO_RESULT, ((char *)"failed to allocate memory"));

    /* "_ellip_harm.pxd":168
 *     if info != 0:
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, size):
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "_ellip_harm.pxd":166
 *              &size, isuppz, work, &lwork, iwork, &liwork, &info)
 * 
 *     if info != 0:             # <<<<<<<<<<<<<<
 *         sf_error.error("ellip_harm", sf_error.NO_RESULT, "failed to allocate memory")
 *         return NULL
 */
  }

  /* "_ellip_harm.pxd":170
 *         return NULL
 * 
 *     for i in range(0, size):             # <<<<<<<<<<<<<<
 *         eigv[i] /= ss[i]
 * 
 */
  __pyx_t_6 = __pyx_v_size;
  __pyx_t_8 = __pyx_t_6;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "_ellip_harm.pxd":171
 * 
 *     for i in range(0, size):
 *         eigv[i] /= ss[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, size):
 */
    __pyx_t_10 = __pyx_v_i;
    (__pyx_v_eigv[__pyx_t_10]) = ((__pyx_v_eigv[__pyx_t_10]) / (__pyx_v_ss[__pyx_v_i]));
  }

  /* "_ellip_harm.pxd":173
 *         eigv[i] /= ss[i]
 * 
 *     for i in range(0, size):             # <<<<<<<<<<<<<<
 *         eigv[i] = eigv[i]/(eigv[size - 1]/pow(-h2, size - 1))
 *     return eigv
 */
  __pyx_t_6 = __pyx_v_size;
  __pyx_t_8 = __pyx_t_6;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "_ellip_harm.pxd":174
 * 
 *     for i in range(0, size):
 *         eigv[i] = eigv[i]/(eigv[size - 1]/pow(-h2, size - 1))             # <<<<<<<<<<<<<<
 *     return eigv
 * 
 */
    (__pyx_v_eigv[__pyx_v_i]) = ((__pyx_v_eigv[__pyx_v_i]) / ((__pyx_v_eigv[(__pyx_v_size - 1)]) / pow((-__pyx_v_h2), (__pyx_v_size - 1))));
  }

  /* "_ellip_harm.pxd":175
 *     for i in range(0, size):
 *         eigv[i] = eigv[i]/(eigv[size - 1]/pow(-h2, size - 1))
 *     return eigv             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_eigv;
  goto __pyx_L0;

  /* "_ellip_harm.pxd":51
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cdef inline double* lame_coefficients(double h2, double k2, int n, int p,             # <<<<<<<<<<<<<<
 *                                       void **bufferp, double signm,
 *                                       double signn) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "_ellip_harm.pxd":181
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cdef inline double ellip_harm_eval(double h2, double k2, int n, int p,             # <<<<<<<<<<<<<<
 *                                    double s, double *eigv, double signm,
 *                                    double signn) nogil:
 */

static CYTHON_INLINE double __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(double __pyx_v_h2, double __pyx_v_k2, int __pyx_v_n, int __pyx_v_p, double __pyx_v_s, double *__pyx_v_eigv, double __pyx_v_signm, double __pyx_v_signn) {
  int __pyx_v_size;
  int __pyx_v_r;
  int __pyx_v_j;
  double __pyx_v_s2;
  double __pyx_v_pp;
  double __pyx_v_lambda_romain;
  double __pyx_v_psi;
  double __pyx_r;
  int __pyx_t_1;
  long __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;

  /* "_ellip_harm.pxd":186
 *     cdef int size, tp, r, j
 *     cdef double s2, pp, lambda_romain, psi
 *     s2 = s*s             # <<<<<<<<<<<<<<
 *     r = n/2
 *     if p - 1 < r + 1:
 */
  __pyx_v_s2 = (__pyx_v_s * __pyx_v_s);

  /* "_ellip_harm.pxd":187
 *     cdef double s2, pp, lambda_romain, psi
 *     s2 = s*s
 *     r = n/2             # <<<<<<<<<<<<<<
 *     if p - 1 < r + 1:
 *         size, psi = r + 1, pow(s, n - 2*r)
 */
  __pyx_v_r = (((long)__pyx_v_n) / 2);

  /* "_ellip_harm.pxd":188
 *     s2 = s*s
 *     r = n/2
 *     if p - 1 < r + 1:             # <<<<<<<<<<<<<<
 *         size, psi = r + 1, pow(s, n - 2*r)
 *     elif p - 1 < (n - r) + (r + 1):
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < (__pyx_v_r + 1)) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":189
 *     r = n/2
 *     if p - 1 < r + 1:
 *         size, psi = r + 1, pow(s, n - 2*r)             # <<<<<<<<<<<<<<
 *     elif p - 1 < (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signm*sqrt(fabs(s2 - h2))
 */
    __pyx_t_2 = (__pyx_v_r + 1);
    __pyx_t_3 = pow(__pyx_v_s, (__pyx_v_n - (2 * __pyx_v_r)));
    __pyx_v_size = __pyx_t_2;
    __pyx_v_psi = __pyx_t_3;

    /* "_ellip_harm.pxd":188
 *     s2 = s*s
 *     r = n/2
 *     if p - 1 < r + 1:             # <<<<<<<<<<<<<<
 *         size, psi = r + 1, pow(s, n - 2*r)
 *     elif p - 1 < (n - r) + (r + 1):
 */
    goto __pyx_L3;
  }

  /* "_ellip_harm.pxd":190
 *     if p - 1 < r + 1:
 *         size, psi = r + 1, pow(s, n - 2*r)
 *     elif p - 1 < (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signm*sqrt(fabs(s2 - h2))
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < ((__pyx_v_n - __pyx_v_r) + (__pyx_v_r + 1))) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":191
 *         size, psi = r + 1, pow(s, n - 2*r)
 *     elif p - 1 < (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signm*sqrt(fabs(s2 - h2))             # <<<<<<<<<<<<<<
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signn*sqrt(fabs(s2 - k2))
 */
    __pyx_t_4 = (__pyx_v_n - __pyx_v_r);
    __pyx_t_3 = ((pow(__pyx_v_s, ((1 - __pyx_v_n) + (2 * __pyx_v_r))) * __pyx_v_signm) * sqrt(fabs((__pyx_v_s2 - __pyx_v_h2))));
    __pyx_v_size = __pyx_t_4;
    __pyx_v_psi = __pyx_t_3;

    /* "_ellip_harm.pxd":190
 *     if p - 1 < r + 1:
 *         size, psi = r + 1, pow(s, n - 2*r)
 *     elif p - 1 < (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signm*sqrt(fabs(s2 - h2))
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 */
    goto __pyx_L3;
  }

  /* "_ellip_harm.pxd":192
 *     elif p - 1 < (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signm*sqrt(fabs(s2 - h2))
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signn*sqrt(fabs(s2 - k2))
 *     elif p - 1 < 2*n + 1:
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < (((__pyx_v_n - __pyx_v_r) + (__pyx_v_n - __pyx_v_r)) + (__pyx_v_r + 1))) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":193
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signm*sqrt(fabs(s2 - h2))
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signn*sqrt(fabs(s2 - k2))             # <<<<<<<<<<<<<<
 *     elif p - 1 < 2*n + 1:
 *         size, psi = r, pow(s,  n - 2*r)*signm*signn*sqrt(fabs((s2 - h2)*(s2 - k2)))
 */
    __pyx_t_4 = (__pyx_v_n - __pyx_v_r);
    __pyx_t_3 = ((pow(__pyx_v_s, ((1 - __pyx_v_n) + (2 * __pyx_v_r))) * __pyx_v_signn) * sqrt(fabs((__pyx_v_s2 - __pyx_v_k2))));
    __pyx_v_size = __pyx_t_4;
    __pyx_v_psi = __pyx_t_3;

    /* "_ellip_harm.pxd":192
 *     elif p - 1 < (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signm*sqrt(fabs(s2 - h2))
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):             # <<<<<<<<<<<<<<
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signn*sqrt(fabs(s2 - k2))
 *     elif p - 1 < 2*n + 1:
 */
    goto __pyx_L3;
  }

  /* "_ellip_harm.pxd":194
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signn*sqrt(fabs(s2 - k2))
 *     elif p - 1 < 2*n + 1:             # <<<<<<<<<<<<<<
 *         size, psi = r, pow(s,  n - 2*r)*signm*signn*sqrt(fabs((s2 - h2)*(s2 - k2)))
 *     lambda_romain = 1.0 - <double>s2/<double>h2
 */
  __pyx_t_1 = (((__pyx_v_p - 1) < ((2 * __pyx_v_n) + 1)) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":195
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signn*sqrt(fabs(s2 - k2))
 *     elif p - 1 < 2*n + 1:
 *         size, psi = r, pow(s,  n - 2*r)*signm*signn*sqrt(fabs((s2 - h2)*(s2 - k2)))             # <<<<<<<<<<<<<<
 *     lambda_romain = 1.0 - <double>s2/<double>h2
 *     pp = eigv[size - 1]
 */
    __pyx_t_4 = __pyx_v_r;
    __pyx_t_3 = (((pow(__pyx_v_s, (__pyx_v_n - (2 * __pyx_v_r))) * __pyx_v_signm) * __pyx_v_signn) * sqrt(fabs(((__pyx_v_s2 - __pyx_v_h2) * (__pyx_v_s2 - __pyx_v_k2)))));
    __pyx_v_size = __pyx_t_4;
    __pyx_v_psi = __pyx_t_3;

    /* "_ellip_harm.pxd":194
 *     elif p - 1 < (n - r) + (n - r) + (r + 1):
 *         size, psi = n - r, pow(s, 1 - n + 2*r)*signn*sqrt(fabs(s2 - k2))
 *     elif p - 1 < 2*n + 1:             # <<<<<<<<<<<<<<
 *         size, psi = r, pow(s,  n - 2*r)*signm*signn*sqrt(fabs((s2 - h2)*(s2 - k2)))
 *     lambda_romain = 1.0 - <double>s2/<double>h2
 */
  }
  __pyx_L3:;

  /* "_ellip_harm.pxd":196
 *     elif p - 1 < 2*n + 1:
 *         size, psi = r, pow(s,  n - 2*r)*signm*signn*sqrt(fabs((s2 - h2)*(s2 - k2)))
 *     lambda_romain = 1.0 - <double>s2/<double>h2             # <<<<<<<<<<<<<<
 *     pp = eigv[size - 1]
 * 
 */
  __pyx_v_lambda_romain = (1.0 - (((double)__pyx_v_s2) / ((double)__pyx_v_h2)));

  /* "_ellip_harm.pxd":197
 *         size, psi = r, pow(s,  n - 2*r)*signm*signn*sqrt(fabs((s2 - h2)*(s2 - k2)))
 *     lambda_romain = 1.0 - <double>s2/<double>h2
 *     pp = eigv[size - 1]             # <<<<<<<<<<<<<<
 * 
 *     for j in range(size - 2, -1, -1):
 */
  __pyx_v_pp = (__pyx_v_eigv[(__pyx_v_size - 1)]);

  /* "_ellip_harm.pxd":199
 *     pp = eigv[size - 1]
 * 
 *     for j in range(size - 2, -1, -1):             # <<<<<<<<<<<<<<
 *         pp = pp*lambda_romain + eigv[j]
 *     pp = pp*psi
 */
  for (__pyx_t_4 = (__pyx_v_size - 2); __pyx_t_4 > -1; __pyx_t_4-=1) {
    __pyx_v_j = __pyx_t_4;

    /* "_ellip_harm.pxd":200
 * 
 *     for j in range(size - 2, -1, -1):
 *         pp = pp*lambda_romain + eigv[j]             # <<<<<<<<<<<<<<
 *     pp = pp*psi
 *     return pp
 */
    __pyx_v_pp = ((__pyx_v_pp * __pyx_v_lambda_romain) + (__pyx_v_eigv[__pyx_v_j]));
  }

  /* "_ellip_harm.pxd":201
 *     for j in range(size - 2, -1, -1):
 *         pp = pp*lambda_romain + eigv[j]
 *     pp = pp*psi             # <<<<<<<<<<<<<<
 *     return pp
 * 
 */
  __pyx_v_pp = (__pyx_v_pp * __pyx_v_psi);

  /* "_ellip_harm.pxd":202
 *         pp = pp*lambda_romain + eigv[j]
 *     pp = pp*psi
 *     return pp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_pp;
  goto __pyx_L0;

  /* "_ellip_harm.pxd":181
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cdef inline double ellip_harm_eval(double h2, double k2, int n, int p,             # <<<<<<<<<<<<<<
 *                                    double s, double *eigv, double signm,
 *                                    double signn) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "_ellip_harm.pxd":205
 * 
 * 
 * cdef inline double ellip_harmonic(double h2, double k2, int n, int p, double s,             # <<<<<<<<<<<<<<
 *                                   double signm, double signn) nogil:
 *     cdef double result
 */

static CYTHON_INLINE double __pyx_f_5scipy_7special_11_ellip_harm_ellip_harmonic(double __pyx_v_h2, double __pyx_v_k2, int __pyx_v_n, int __pyx_v_p, double __pyx_v_s, double __pyx_v_signm, double __pyx_v_signn) {
  double __pyx_v_result;
  double *__pyx_v_eigv;
  void *__pyx_v_bufferp;
  double __pyx_r;
  int __pyx_t_1;

  /* "_ellip_harm.pxd":210
 *     cdef double *eigv
 *     cdef void *bufferp
 *     eigv = lame_coefficients(h2, k2, n, p, &bufferp, signm, signn)             # <<<<<<<<<<<<<<
 *     if not eigv:
 *         free(bufferp)
 */
  __pyx_v_eigv = __pyx_f_5scipy_7special_11_ellip_harm_lame_coefficients(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, (&__pyx_v_bufferp), __pyx_v_signm, __pyx_v_signn);

  /* "_ellip_harm.pxd":211
 *     cdef void *bufferp
 *     eigv = lame_coefficients(h2, k2, n, p, &bufferp, signm, signn)
 *     if not eigv:             # <<<<<<<<<<<<<<
 *         free(bufferp)
 *         return NAN
 */
  __pyx_t_1 = ((!(__pyx_v_eigv != 0)) != 0);
  if (__pyx_t_1) {

    /* "_ellip_harm.pxd":212
 *     eigv = lame_coefficients(h2, k2, n, p, &bufferp, signm, signn)
 *     if not eigv:
 *         free(bufferp)             # <<<<<<<<<<<<<<
 *         return NAN
 *     result = ellip_harm_eval(h2, k2, n, p, s, eigv, signm, signn)
 */
    free(__pyx_v_bufferp);

    /* "_ellip_harm.pxd":213
 *     if not eigv:
 *         free(bufferp)
 *         return NAN             # <<<<<<<<<<<<<<
 *     result = ellip_harm_eval(h2, k2, n, p, s, eigv, signm, signn)
 *     free(bufferp)
 */
    __pyx_r = NPY_NAN;
    goto __pyx_L0;

    /* "_ellip_harm.pxd":211
 *     cdef void *bufferp
 *     eigv = lame_coefficients(h2, k2, n, p, &bufferp, signm, signn)
 *     if not eigv:             # <<<<<<<<<<<<<<
 *         free(bufferp)
 *         return NAN
 */
  }

  /* "_ellip_harm.pxd":214
 *         free(bufferp)
 *         return NAN
 *     result = ellip_harm_eval(h2, k2, n, p, s, eigv, signm, signn)             # <<<<<<<<<<<<<<
 *     free(bufferp)
 *     return result
 */
  __pyx_v_result = __pyx_f_5scipy_7special_11_ellip_harm_ellip_harm_eval(__pyx_v_h2, __pyx_v_k2, __pyx_v_n, __pyx_v_p, __pyx_v_s, __pyx_v_eigv, __pyx_v_signm, __pyx_v_signn);

  /* "_ellip_harm.pxd":215
 *         return NAN
 *     result = ellip_harm_eval(h2, k2, n, p, s, eigv, signm, signn)
 *     free(bufferp)             # <<<<<<<<<<<<<<
 *     return result
 */
  free(__pyx_v_bufferp);

  /* "_ellip_harm.pxd":216
 *     result = ellip_harm_eval(h2, k2, n, p, s, eigv, signm, signn)
 *     free(bufferp)
 *     return result             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "_ellip_harm.pxd":205
 * 
 * 
 * cdef inline double ellip_harmonic(double h2, double k2, int n, int p, double s,             # <<<<<<<<<<<<<<
 *                                   double signm, double signn) nogil:
 *     cdef double result
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":868
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":869
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":870
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":868
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":872
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":873
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":874
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":875
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":874
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":876
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":872
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":880
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":882
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 882, __pyx_L3_error)

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":883
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 883, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":884
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 884, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 884, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":881
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":880
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":886
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":888
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 888, __pyx_L3_error)

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":889
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 889, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":890
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 890, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 890, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":887
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":886
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":894
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 894, __pyx_L3_error)

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":895
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 895, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":896
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 896, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 896, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":893
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__ellip_harm_2(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__ellip_harm_2},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_ellip_harm_2",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 1},
  {&__pyx_n_u_F_integrand, __pyx_k_F_integrand, sizeof(__pyx_k_F_integrand), 0, 1, 0, 1},
  {&__pyx_n_u_F_integrand1, __pyx_k_F_integrand1, sizeof(__pyx_k_F_integrand1), 0, 1, 0, 1},
  {&__pyx_n_u_F_integrand2, __pyx_k_F_integrand2, sizeof(__pyx_k_F_integrand2), 0, 1, 0, 1},
  {&__pyx_n_u_F_integrand3, __pyx_k_F_integrand3, sizeof(__pyx_k_F_integrand3), 0, 1, 0, 1},
  {&__pyx_n_u_F_integrand4, __pyx_k_F_integrand4, sizeof(__pyx_k_F_integrand4), 0, 1, 0, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_LowLevelCallable, __pyx_k_LowLevelCallable, sizeof(__pyx_k_LowLevelCallable), 0, 0, 1, 1},
  {&__pyx_n_u_alg, __pyx_k_alg, sizeof(__pyx_k_alg), 0, 1, 0, 1},
  {&__pyx_n_s_bufferp, __pyx_k_bufferp, sizeof(__pyx_k_bufferp), 0, 0, 1, 1},
  {&__pyx_n_s_capsule, __pyx_k_capsule, sizeof(__pyx_k_capsule), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_ctypes, __pyx_k_ctypes, sizeof(__pyx_k_ctypes), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_eigv, __pyx_k_eigv, sizeof(__pyx_k_eigv), 0, 0, 1, 1},
  {&__pyx_kp_s_ellip_harm_2_pyx, __pyx_k_ellip_harm_2_pyx, sizeof(__pyx_k_ellip_harm_2_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_ellipsoid, __pyx_k_ellipsoid, sizeof(__pyx_k_ellipsoid), 0, 0, 1, 1},
  {&__pyx_n_s_ellipsoid_norm, __pyx_k_ellipsoid_norm, sizeof(__pyx_k_ellipsoid_norm), 0, 0, 1, 1},
  {&__pyx_n_s_epsabs, __pyx_k_epsabs, sizeof(__pyx_k_epsabs), 0, 0, 1, 1},
  {&__pyx_n_s_epsrel, __pyx_k_epsrel, sizeof(__pyx_k_epsrel), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_err1, __pyx_k_err1, sizeof(__pyx_k_err1), 0, 0, 1, 1},
  {&__pyx_n_s_err2, __pyx_k_err2, sizeof(__pyx_k_err2), 0, 0, 1, 1},
  {&__pyx_n_s_err3, __pyx_k_err3, sizeof(__pyx_k_err3), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_eval, __pyx_k_eval, sizeof(__pyx_k_eval), 0, 0, 1, 1},
  {&__pyx_n_s_from_cython, __pyx_k_from_cython, sizeof(__pyx_k_from_cython), 0, 0, 1, 1},
  {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
  {&__pyx_n_s_h2, __pyx_k_h2, sizeof(__pyx_k_h2), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_k2, __pyx_k_k2, sizeof(__pyx_k_k2), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mod, __pyx_k_mod, sizeof(__pyx_k_mod), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_quad, __pyx_k_quad, sizeof(__pyx_k_quad), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_res1, __pyx_k_res1, sizeof(__pyx_k_res1), 0, 0, 1, 1},
  {&__pyx_n_s_res2, __pyx_k_res2, sizeof(__pyx_k_res2), 0, 0, 1, 1},
  {&__pyx_n_s_res3, __pyx_k_res3, sizeof(__pyx_k_res3), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_scipy__lib__ccallback, __pyx_k_scipy__lib__ccallback, sizeof(__pyx_k_scipy__lib__ccallback), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_integrate, __pyx_k_scipy_integrate, sizeof(__pyx_k_scipy_integrate), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_special__ellip_harm_2, __pyx_k_scipy_special__ellip_harm_2, sizeof(__pyx_k_scipy_special__ellip_harm_2), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_weight, __pyx_k_weight, sizeof(__pyx_k_weight), 0, 0, 1, 1},
  {&__pyx_n_s_wvar, __pyx_k_wvar, sizeof(__pyx_k_wvar), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 115, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 884, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "scipy/special/_ellip_harm_2.pyx":162
 *         capsule = PyCapsule_New(<void*>&data, NULL, NULL)
 * 
 *         wvar = (-0.5, -0.5)             # <<<<<<<<<<<<<<
 * 
 *         res, err = quad(LowLevelCallable.from_cython(mod, "_F_integrand1", capsule), h, k,
 */
  __pyx_tuple__2 = PyTuple_Pack(2, __pyx_float_neg_0_5, __pyx_float_neg_0_5); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "scipy/special/_ellip_harm_2.pyx":170
 *                           epsabs=1e-300, epsrel=1e-15, weight="alg", wvar=wvar)
 * 
 *         wvar = (0, -0.5)             # <<<<<<<<<<<<<<
 * 
 *         res2, err2 = quad(LowLevelCallable.from_cython(mod, "_F_integrand3", capsule), 0, h,
 */
  __pyx_tuple__3 = PyTuple_Pack(2, __pyx_int_0, __pyx_float_neg_0_5); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":884
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":890
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(2, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "scipy/special/_ellip_harm_2.pyx":105
 * 
 * 
 * def _ellipsoid(double h2, double k2, int n, int p, double s):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */
  __pyx_tuple__6 = PyTuple_Pack(13, __pyx_n_s_h2, __pyx_n_s_k2, __pyx_n_s_n, __pyx_n_s_p, __pyx_n_s_s, __pyx_n_s_mod, __pyx_n_s_quad, __pyx_n_s_data, __pyx_n_s_eval, __pyx_n_s_bufferp, __pyx_n_s_res, __pyx_n_s_err, __pyx_n_s_capsule); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(5, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ellip_harm_2_pyx, __pyx_n_s_ellipsoid, 105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 105, __pyx_L1_error)

  /* "scipy/special/_ellip_harm_2.pyx":137
 * 
 * 
 * def _ellipsoid_norm(double h2, double k2, int n, int p):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */
  __pyx_tuple__8 = PyTuple_Pack(24, __pyx_n_s_h2, __pyx_n_s_k2, __pyx_n_s_n, __pyx_n_s_p, __pyx_n_s_mod, __pyx_n_s_quad, __pyx_n_s_data, __pyx_n_s_eigv, __pyx_n_s_bufferp, __pyx_n_s_eval, __pyx_n_s_res, __pyx_n_s_res1, __pyx_n_s_res2, __pyx_n_s_res3, __pyx_n_s_err, __pyx_n_s_err1, __pyx_n_s_err2, __pyx_n_s_err3, __pyx_n_s_h, __pyx_n_s_k, __pyx_n_s_capsule, __pyx_n_s_wvar, __pyx_n_s_error, __pyx_n_s_result); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(4, 0, 24, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ellip_harm_2_pyx, __pyx_n_s_ellipsoid_norm, 137, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_1eneg_15 = PyFloat_FromDouble(1e-15); if (unlikely(!__pyx_float_1eneg_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_300 = PyFloat_FromDouble(1e-300); if (unlikely(!__pyx_float_1eneg_300)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_0_5 = PyFloat_FromDouble(-0.5); if (unlikely(!__pyx_float_neg_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("_F_integrand", (void (*)(void))__pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand, "double (double, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_F_integrand1", (void (*)(void))__pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand1, "double (double, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_F_integrand2", (void (*)(void))__pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand2, "double (double, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_F_integrand3", (void (*)(void))__pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand3, "double (double, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_F_integrand4", (void (*)(void))__pyx_f_5scipy_7special_13_ellip_harm_2__F_integrand4, "double (double, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_ellip_harm_2(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_ellip_harm_2(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__ellip_harm_2(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__ellip_harm_2(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__ellip_harm_2(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_ellip_harm_2' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__ellip_harm_2(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_ellip_harm_2", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_scipy__special___ellip_harm_2) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "scipy.special._ellip_harm_2")) {
      if (unlikely(PyDict_SetItemString(modules, "scipy.special._ellip_harm_2", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code();
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "scipy/special/_ellip_harm_2.pyx":1
 * import ctypes             # <<<<<<<<<<<<<<
 * from libc.math cimport sqrt, fabs
 * from libc.stdlib cimport free
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_ctypes, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctypes, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":4
 * from libc.math cimport sqrt, fabs
 * from libc.stdlib cimport free
 * from numpy import nan             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "Python.h":
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_nan);
  __Pyx_GIVEREF(__pyx_n_s_nan);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_nan);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nan, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":9
 *     object PyCapsule_New(void *pointer, char *name, void *destructor)
 * 
 * from scipy._lib._ccallback import LowLevelCallable             # <<<<<<<<<<<<<<
 * from ._ellip_harm cimport ellip_harmonic, ellip_harm_eval, lame_coefficients
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_LowLevelCallable);
  __Pyx_GIVEREF(__pyx_n_s_LowLevelCallable);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_LowLevelCallable);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_scipy__lib__ccallback, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_LowLevelCallable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LowLevelCallable, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":105
 * 
 * 
 * def _ellipsoid(double h2, double k2, int n, int p, double s):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5scipy_7special_13_ellip_harm_2_1_ellipsoid, NULL, __pyx_n_s_scipy_special__ellip_harm_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ellipsoid, __pyx_t_1) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":137
 * 
 * 
 * def _ellipsoid_norm(double h2, double k2, int n, int p):             # <<<<<<<<<<<<<<
 *     import scipy.special._ellip_harm_2 as mod
 *     from scipy.integrate import quad
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5scipy_7special_13_ellip_harm_2_3_ellipsoid_norm, NULL, __pyx_n_s_scipy_special__ellip_harm_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ellipsoid_norm, __pyx_t_1) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "scipy/special/_ellip_harm_2.pyx":193
 * cimport numpy as np
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * np.import_ufunc()
 * 
 */
  __pyx_t_3 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 193, __pyx_L1_error)

  /* "scipy/special/_ellip_harm_2.pyx":194
 * 
 * np.import_array()
 * np.import_ufunc()             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "numpy/ufuncobject.h":
 */
  __pyx_t_3 = __pyx_f_5numpy_import_ufunc(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 194, __pyx_L1_error)

  /* "scipy/special/_ellip_harm_2.pyx":1
 * import ctypes             # <<<<<<<<<<<<<<
 * from libc.math cimport sqrt, fabs
 * from libc.stdlib cimport free
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../venv_py38/lib/python3.8/site-packages/numpy/__init__.pxd":892
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init scipy.special._ellip_harm_2", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init scipy.special._ellip_harm_2");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* None */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_CBLAS_INT(CBLAS_INT value) {
    const CBLAS_INT neg_one = (CBLAS_INT) ((CBLAS_INT) 0 - (CBLAS_INT) 1), const_zero = (CBLAS_INT) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(CBLAS_INT) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(CBLAS_INT) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(CBLAS_INT) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(CBLAS_INT) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(CBLAS_INT) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(CBLAS_INT),
                                     little, !is_unsigned);
    }
}

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE CBLAS_INT __Pyx_PyInt_As_CBLAS_INT(PyObject *x) {
    const CBLAS_INT neg_one = (CBLAS_INT) ((CBLAS_INT) 0 - (CBLAS_INT) 1), const_zero = (CBLAS_INT) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(CBLAS_INT) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(CBLAS_INT, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (CBLAS_INT) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (CBLAS_INT) 0;
                case  1: __PYX_VERIFY_RETURN_INT(CBLAS_INT, digit, digits[0])
                case 2:
                    if (8 * sizeof(CBLAS_INT) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) >= 2 * PyLong_SHIFT) {
                            return (CBLAS_INT) (((((CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(CBLAS_INT) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) >= 3 * PyLong_SHIFT) {
                            return (CBLAS_INT) (((((((CBLAS_INT)digits[2]) << PyLong_SHIFT) | (CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(CBLAS_INT) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) >= 4 * PyLong_SHIFT) {
                            return (CBLAS_INT) (((((((((CBLAS_INT)digits[3]) << PyLong_SHIFT) | (CBLAS_INT)digits[2]) << PyLong_SHIFT) | (CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (CBLAS_INT) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(CBLAS_INT) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(CBLAS_INT, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(CBLAS_INT) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(CBLAS_INT, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (CBLAS_INT) 0;
                case -1: __PYX_VERIFY_RETURN_INT(CBLAS_INT, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(CBLAS_INT,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(CBLAS_INT) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) - 1 > 2 * PyLong_SHIFT) {
                            return (CBLAS_INT) (((CBLAS_INT)-1)*(((((CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(CBLAS_INT) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) - 1 > 2 * PyLong_SHIFT) {
                            return (CBLAS_INT) ((((((CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(CBLAS_INT) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) - 1 > 3 * PyLong_SHIFT) {
                            return (CBLAS_INT) (((CBLAS_INT)-1)*(((((((CBLAS_INT)digits[2]) << PyLong_SHIFT) | (CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(CBLAS_INT) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) - 1 > 3 * PyLong_SHIFT) {
                            return (CBLAS_INT) ((((((((CBLAS_INT)digits[2]) << PyLong_SHIFT) | (CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(CBLAS_INT) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) - 1 > 4 * PyLong_SHIFT) {
                            return (CBLAS_INT) (((CBLAS_INT)-1)*(((((((((CBLAS_INT)digits[3]) << PyLong_SHIFT) | (CBLAS_INT)digits[2]) << PyLong_SHIFT) | (CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(CBLAS_INT) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(CBLAS_INT, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(CBLAS_INT) - 1 > 4 * PyLong_SHIFT) {
                            return (CBLAS_INT) ((((((((((CBLAS_INT)digits[3]) << PyLong_SHIFT) | (CBLAS_INT)digits[2]) << PyLong_SHIFT) | (CBLAS_INT)digits[1]) << PyLong_SHIFT) | (CBLAS_INT)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(CBLAS_INT) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(CBLAS_INT, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(CBLAS_INT) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(CBLAS_INT, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            CBLAS_INT val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (CBLAS_INT) -1;
        }
    } else {
        CBLAS_INT val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (CBLAS_INT) -1;
        val = __Pyx_PyInt_As_CBLAS_INT(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to CBLAS_INT");
    return (CBLAS_INT) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to CBLAS_INT");
    return (CBLAS_INT) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
